# Copyright: (c) 2018, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import json
import logging
import logging.config
import os

try:
    from logging import NullHandler
except ImportError:  # pragma: no cover
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass


def _setup_logging(logger):
    log_path = os.environ.get('PYPSRP_LOG_CFG', None)

    if log_path is not None and os.path.exists(log_path):  # pragma: no cover
        # log log config from JSON file
        with open(log_path, 'rt') as f:
            config = json.load(f)

        logging.config.dictConfig(config)
    else:
        # no logging was provided
        logger.addHandler(NullHandler())


logger = logging.getLogger(__name__)
_setup_logging(logger)

# Contains a list of features, used by external libraries to determine whether
# a new enough pypsrp is installed to support the features it needs
FEATURES = [
    'wsman_locale',
    'wsman_read_timeout',
    'wsman_reconnections',
]



# Copyright: (c) 2018, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from pypsrp.complex_objects import PSInvocationState, RunspacePoolState


class WinRMError(Exception):
    # Base WinRM Error
    pass


class AuthenticationError(WinRMError):
    # Used when the user failed to authenticate
    pass


class WinRMTransportError(WinRMError):
    # An error occurred during the transport stage

    @property
    def protocol(self):
        return self.args[0]

    @property
    def code(self):
        return self.args[1]

    @property
    def response_text(self):
        return self.args[2]

    @property
    def message(self):
        return "Bad %s response returned from the server. Code: %d, " \
               "Content: '%s'"\
               % (self.protocol.upper(), self.code, self.response_text)

    def __str__(self):
        return self.message


class WSManFaultError(WinRMError):
    # Contains the WSManFault information if a WSManFault was received

    @property
    def code(self):
        return self.args[0]

    @property
    def machine(self):
        return self.args[1]

    @property
    def reason(self):
        return self.args[2]

    @property
    def provider(self):
        return self.args[3]

    @property
    def provider_path(self):
        return self.args[4]

    @property
    def provider_fault(self):
        return self.args[5]

    @property
    def message(self):
        error_details = []
        if self.code:
            error_details.append("Code: %s" % self.code)

        if self.machine:
            error_details.append("Machine: %s" % self.machine)

        if self.reason:
            error_details.append("Reason: %s" % self.reason)

        if self.provider:
            error_details.append("Provider: %s" % self.provider)

        if self.provider_path:
            error_details.append("Provider Path: %s" % self.provider_path)

        if self.provider_fault:
            error_details.append("Provider Fault: %s" % self.provider_fault)

        if len(error_details) == 0:
            error_details.append("No details returned by the server")

        error_msg = "Received a WSManFault message. (%s)" \
                    % ", ".join(error_details)
        return error_msg

    def __str__(self):
        return self.message


# PSRP Exceptions below
class _InvalidStateError(WinRMError):
    _STATE_OBJ = None

    @property
    def current_state(self):
        return self.args[0]

    @property
    def expected_state(self):
        return self.args[1]

    @property
    def action(self):
        return self.args[2]

    @property
    def message(self):
        current_state = str(self._STATE_OBJ(self.current_state))
        expected_state = self.expected_state
        if not isinstance(expected_state, list):
            expected_state = [expected_state]
        exp_state = [str(self._STATE_OBJ(s)) for s in expected_state]
        return "Cannot '%s' on the current state '%s', expecting state(s): " \
               "'%s'" % (self.action, current_state, ", ".join(exp_state))

    def __str__(self):
        return self.message


class InvalidRunspacePoolStateError(_InvalidStateError):
    # Used in PSRP when the state of a RunspacePool does not meet the required
    # state for the operation to run
    _STATE_OBJ = RunspacePoolState


class InvalidPipelineStateError(_InvalidStateError):
    # Used in PSRP when teh state of a PowerShell Pipeline does not meet the
    # required state for the operation to run
    _STATE_OBJ = PSInvocationState


class InvalidPSRPOperation(WinRMError):
    # Generic error used to denote an operation that is invalid or could not
    # run until other conditions are met
    pass


class FragmentError(WinRMError):
    # Any error occurred during the packet fragmentation
    pass


class SerializationError(WinRMError):
    # Any error during the serialization process
    pass



# Copyright: (c) 2018, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import base64
import binascii
import logging
import re
import sys
import uuid

from copy import copy
from cryptography.hazmat.primitives.padding import PKCS7
from six import string_types

from pypsrp.complex_objects import ApartmentState, Color, \
    CommandMetadataCount, CommandOrigin, Coordinates, ComplexObject, \
    CultureInfo, DictionaryMeta, GenericComplexObject, HostMethodIdentifier, \
    InformationalRecord, KeyInfoDotNet, ListMeta, ObjectMeta, \
    ParameterMetadata, PipelineResultTypes, ProgressRecordType, PSCredential,\
    PSThreadOptions, QueueMeta, RemoteStreamOptions, \
    SessionStateEntryVisibility, Size, StackMeta
from pypsrp.exceptions import SerializationError
from pypsrp.messages import DebugRecord, ErrorRecord, InformationRecord, \
    VerboseRecord, WarningRecord
from pypsrp._utils import to_bytes, to_string, to_unicode

try:
    from queue import Queue, Empty
except ImportError:  # pragma: no cover
    from Queue import Queue, Empty

if sys.version_info[0] == 2 and sys.version_info[1] < 7:  # pragma: no cover
    # ElementTree in Python 2.6 does not support namespaces so we need to use
    # lxml instead for this version
    from lxml import etree as ET
    element_type = ET._Element
else:  # pragma: no cover
    import xml.etree.ElementTree as ET
    element_type = ET.Element

log = logging.getLogger(__name__)


class Serializer(object):

    def __init__(self):
        self.obj_id = 0
        self.obj = {}
        self.tn_id = 0
        self.tn = {}

        self.cipher = None
        # Finds C0, C1 and surrogate pairs in a unicode string for us to
        # encode according to the PSRP rules
        if sys.maxunicode == 65535:  # pragma: no cover
            # using a narrow Python build or Python 2.x, the regex we need to
            # use to find surrogate pairs is different than a wide build or on
            # Python 3
            self._serial_str = re.compile(u"[\u0000-\u001F]|"
                                          u"[\u007F-\u009F]|"
                                          u"[\uD800-\uDBFF][\uDC00-\uDFFF]")
        else:  # pragma: no cover
            self._serial_str = re.compile(u'[\u0000-\u001F'
                                          u'\u007F-\u009F'
                                          u'\U00010000-\U0010FFFF]')

        # to support surrogate UTF-16 pairs we need to use a UTF-16 regex
        # so we can replace the UTF-16 string representation with the actual
        # UTF-16 byte value and then decode that
        self._deserial_str = re.compile(b"\\x00_\\x00x([\\0\\w]{8})\\x00_")

    def serialize(self, value, metadata=None, parent=None, clear=True):
        """
        Serializes a raw value or class into an XML Element that can be sent
        over to the remote host.

        :param value: The value to serialize
        :param metadata: Any extra metadata to control how to serialize the
            value, if None then the value will be inferred by the type
        :param parent: Whether to append the element onto a parent element
        :param clear: Whether to clear the Obj and TN reference map, this
            should only be True when initially calling serialize
        :return: The XML Element from the serializied value
        """
        if clear:
            self._clear()

        if isinstance(value, element_type):
            if metadata is not None and metadata.name is not None:
                value.attrib['N'] = metadata.name

            if parent is not None:
                parent.append(value)

            return value

        metadata = metadata or ObjectMeta()
        if metadata.tag == "*":
            metadata.tag = self._get_tag_from_value(value)

        pack_function = {
            # primitive types
            'S': lambda m, d: self._serialize_string(d),
            'ToString': lambda d: self._serialize_string(d),
            'C': lambda m, d: str(ord(d)),
            'B': lambda m, d: str(d).lower(),
            'DT': lambda m, d: None,
            'TS': lambda m, d: str(d),
            'By': lambda m, d: str(d),
            'SB': lambda m, d: str(d),
            'U16': lambda m, d: str(d),
            'I16': lambda m, d: str(d),
            'U32': lambda m, d: str(d),
            'I32': lambda m, d: str(d),
            'U64': lambda m, d: str(d),
            'I64': lambda m, d: str(d),
            'Sg': lambda m, d: str(d),
            'Db': lambda m, d: str(d),
            'D': lambda m, d: str(d),
            'BA': lambda m, d: to_string(base64.b64encode(d)),
            'G': lambda m, d: str(d),
            'URI': lambda m, d: self._serialize_string(d),
            'Version': lambda m, d: str(d),
            'XD': lambda m, d: self._serialize_string(d),
            'SBK': lambda m, d: self._serialize_string(d),
            'SS': lambda m, d: self._serialize_secure_string(d),
            'Obj': self._serialize_obj,
            "ObjDynamic": self._serialize_dynamic_obj,
            'LST': self._serialize_lst,
            'IE': self._serialize_ie,
            'QUE': self._serialize_que,
            'STK': self._serialize_stk,
            'DCT': self._serialize_dct
        }[metadata.tag]

        if value is None:
            if metadata.optional:
                return
            element = ET.Element("Nil")
        else:
            element_value = pack_function(metadata, value)
            if isinstance(element_value, string_types):
                element = ET.Element(metadata.tag)
                element.text = element_value
            else:
                element = element_value

        if metadata.name is not None:
            element.attrib['N'] = metadata.name

        if parent is not None:
            parent.append(element)

        return element

    def deserialize(self, element, metadata=None, clear=True):
        if clear:
            self._clear()

        if isinstance(element, string_types):
            element_string = element
            try:
                element = ET.fromstring(element)
            except ET.ParseError as err:
                log.warning("Failed to parse data '%s' as XML, return raw "
                            "xml: %s" % (element_string, str(err)))
                return element_string
        else:
            xml_string = ET.tostring(element, encoding='utf-8', method='xml')
            element_string = to_string(xml_string)

        metadata = metadata or ObjectMeta()
        if metadata.tag == "*":
            metadata.tag = element.tag

        # get the object types so we store the TN Ref ids for later use
        obj_types = self._get_types_from_obj(element)

        # check if it is a primitive object
        unpack_function = {
            # Primitive types
            'S': lambda d: self._deserialize_string(d.text),
            'ToString': lambda d: self._deserialize_string(d.text),
            'C': lambda d: chr(int(d.text)),
            'B': lambda d: d.text.lower() == "true",
            'DT': lambda d: d.text,
            'TS': lambda d: d.text,
            'By': lambda d: int(d.text),
            'SB': lambda d: int(d.text),
            'U16': lambda d: int(d.text),
            'I16': lambda d: int(d.text),
            'U32': lambda d: int(d.text),
            'I32': lambda d: int(d.text),
            'U64': lambda d: int(d.text),
            'I64': lambda d: int(d.text),
            'Sg': lambda d: float(d.text),
            'Db': lambda d: float(d.text),
            'D': lambda d: d.text,  # TODO: deserialize this
            'BA': lambda d: base64.b64decode(d.text),
            'G': lambda d: uuid.UUID(d.text),
            'URI': lambda d: self._deserialize_string(d.text),
            'Nil': lambda d: None,
            'Version': lambda d: d.text,
            'XD': lambda d: self._deserialize_string(d.text),
            'SBK': lambda d: self._deserialize_string(d.text),
            'SS': lambda d: self._deserialize_secure_string(d),

            # references an object already deserialized
            'Ref': lambda d: self.obj[d.attrib['RefId']],
        }.get(element.tag)

        if unpack_function is not None:
            return unpack_function(element)

        # not a primitive object, so try and decode the complex object
        if type(metadata) == ObjectMeta and metadata.object is None:
            structures = {
                "Selected.Microsoft.PowerShell.Commands.GenericMeasureInfo":
                    ObjectMeta("Obj", object=CommandMetadataCount),
                "System.Array": ListMeta(),
                "System.Collections.ArrayList": ListMeta(),
                "System.Collections.Hashtable": DictionaryMeta(),
                "System.Collections.Generic.List": ListMeta(),
                "System.Collections.Queue": QueueMeta(),
                "System.Collections.Stack": StackMeta(),
                "System.ConsoleColor": ObjectMeta("Obj", object=Color),
                "System.Management.Automation.CommandOrigin":
                    ObjectMeta("Obj", object=CommandOrigin),
                "System.Management.Automation.DebugRecord":
                    ObjectMeta("Obj", object=DebugRecord),
                "System.Management.Automation.ErrorRecord":
                    ObjectMeta("Obj", object=ErrorRecord),
                "System.Management.Automation.Host.Coordinates":
                    ObjectMeta("Obj", object=Coordinates),
                "System.Management.Automation.Host.KeyInfo":
                    ObjectMeta("Obj", object=KeyInfoDotNet),
                "System.Management.Automation.Host.Size":
                    ObjectMeta("Obj", object=Size),
                "System.Management.Automation.InformationalRecord":
                    ObjectMeta("Obj", object=InformationalRecord),
                "System.Management.Automation.InformationRecord":
                    ObjectMeta("Obj", object=InformationRecord),
                "System.Management.Automation.ParameterMetadata":
                    ObjectMeta("Obj", object=ParameterMetadata),
                "System.Management.Automation.ProgressRecordType":
                    ObjectMeta("Obj", object=ProgressRecordType),
                "System.Management.Automation.PSBoundParametersDictionary":
                    DictionaryMeta(),
                "System.Management.Automation.PSCredential":
                    ObjectMeta("Obj", object=PSCredential),
                "System.Management.Automation.PSObject":
                    ObjectMeta("ObjDynamic", object=GenericComplexObject),
                "System.Management.Automation.PSPrimitiveDictionary":
                    DictionaryMeta(),
                "System.Management.Automation.PSTypeName": ObjectMeta("S"),
                "System.Management.Automation.Remoting.RemoteHostMethodId":
                    ObjectMeta("Obj", object=HostMethodIdentifier),
                "System.Management.Automation.Runspaces.ApartmentState":
                    ObjectMeta("Obj", object=ApartmentState),
                "System.Management.Automation.Runspaces.PipelineResultTypes":
                    ObjectMeta("Obj", object=PipelineResultTypes),
                "System.Management.Automation.Runspaces.PSThreadOptions":
                    ObjectMeta("Obj", object=PSThreadOptions),
                "System.Management.Automation.Runspaces.RemoteStreamOptions":
                    ObjectMeta("Obj", object=RemoteStreamOptions),
                "System.Management.Automation.SessionStateEntryVisibility":
                    ObjectMeta("Obj", object=SessionStateEntryVisibility),
                "System.Management.Automation.VerboseRecord":
                    ObjectMeta("Obj", object=VerboseRecord),
                "System.Management.Automation.WarningRecord":
                    ObjectMeta("Obj", object=WarningRecord),
                "System.Globalization.CultureInfo":
                    ObjectMeta("Obj", object=CultureInfo),

                # Fallback to the GenericComplexObject
                "System.Object":
                    ObjectMeta("ObjDynamic", object=GenericComplexObject),

                # Primitive types
                "System.String": ObjectMeta("S"),
                "System.Char": ObjectMeta("C"),
                "System.Boolean": ObjectMeta("B"),
                "System.DateTime": ObjectMeta("DT"),
                # None: ObjectMeta("TS"), # duration timespan
                "System.Byte": ObjectMeta("By"),
                "System.SByte": ObjectMeta("SB"),
                "System.UInt16": ObjectMeta("U16"),
                "System.Int16": ObjectMeta("I16"),
                "System.UInt32": ObjectMeta("U32"),
                "System.Int32": ObjectMeta("I32"),
                "System.UInt64": ObjectMeta("U64"),
                "System.Int64": ObjectMeta("I64"),
                "System.Single": ObjectMeta("Sg"),
                "System.Double": ObjectMeta("Db"),
                "System.Decimal": ObjectMeta("D"),
                # None: ObjectMeta("BA"), # Byte array base64 encoded
                "System.Guid": ObjectMeta("G"),
                "System.Uri": ObjectMeta("URI"),
                "System.Version": ObjectMeta("Version"),
                "System.Xml.XmlDocument": ObjectMeta("XD"),
                "System.Management.Automation.ScriptBlock": ObjectMeta("SBK"),
                "System.Security.SecureString": ObjectMeta("SS"),
            }

            # fallback to GenericComplexObject if no types were defined
            if metadata.tag == "Obj" and len(obj_types) == 0:
                obj_types = ["System.Object"]

            metadata = None
            for obj_type in obj_types:
                if obj_type.startswith("Deserialized.System."):
                    obj_type = obj_type[13:]

                is_list = False
                if obj_type.endswith("[]"):
                    obj_type = obj_type[0:-2]
                    is_list = True
                elif obj_type.startswith("System.Collections."
                                         "Generic.List`1[["):
                    list_info = obj_type[35:-1]
                    obj_type = list_info.split(",")[0]
                    is_list = True
                elif obj_type.startswith("System.Collections.ObjectModel."
                                         "Collection`1[["):
                    list_info = obj_type[45:-1]
                    obj_type = list_info.split(",")[0]
                    is_list = True
                elif obj_type.startswith("System.Collections.ObjectModel."
                                         "ReadOnlyCollection`1[["):
                    list_info = obj_type[53:-1]
                    obj_type = list_info.split(",")[0]
                    is_list = True
                elif obj_type.startswith("System.Collections.Generic."
                                         "Dictionary`2[["):
                    dict_meta = obj_type[41:-2].split("],[")
                    key_type = structures.get(dict_meta[0].split(",")[0],
                                              ObjectMeta())
                    value_type = structures.get(dict_meta[1].split(",")[0],
                                                ObjectMeta())
                    metadata = DictionaryMeta(dict_key_meta=key_type,
                                              dict_value_meta=value_type)
                    break

                obj_meta = structures.get(obj_type)
                if obj_meta is not None:
                    metadata = obj_meta
                    if is_list:
                        metadata = ListMeta(list_value_meta=metadata)
                    break

        # we were unable to find the complex object type so just return the
        # element
        if metadata is None:
            obj = element_string
        elif metadata.tag == "Obj":
            obj = self._deserialize_obj(element, metadata)
        elif metadata.tag == "ObjDynamic":
            obj = self._deserialize_dynamic_obj(element, metadata)
        elif metadata.tag == "LST":
            obj = self._deserialize_lst(element, metadata)
        elif metadata.tag == "QUE":
            obj = self._deserialize_que(element)
        elif metadata.tag == "STK":
            obj = self._deserialize_stk(element)
        elif metadata.tag == "DCT":
            obj = self._deserialize_dct(element)
        else:
            log.warning("Unknown metadata tag type '%s', failed to "
                        "deserialize object" % metadata.tag)
            obj = element_string

        if element.tag == "Obj":
            self.obj[element.attrib['RefId']] = obj

        if isinstance(obj, ComplexObject):
            obj._xml = element_string

        return obj

    def _get_tag_from_value(self, value):
        # Get's the XML tag based on the value type, this is a simple list
        # and explicit tagging is recommended.

        value_type = type(value)
        if value_type == int:
            return "I32"
        elif value_type == bool:
            return "B"
        elif value_type == float:
            return "Sg"
        elif value_type == str:
            return "S"
        elif value_type == bytes:
            # This will only occur in Python 3 as a byte string in Python 2 is
            # a str. If users on that platform want a BA then they need to
            # explicitly set the metadata themselves
            return "BA"
        elif value_type == uuid.UUID:
            return "G"
        elif value_type == list:
            return "LST"
        elif value_type == dict:
            return "DCT"
        elif isinstance(value, Queue):
            return "QUE"
        elif isinstance(value, GenericComplexObject):
            return "ObjDynamic"
        elif isinstance(value, ComplexObject):
            return "Obj"
        else:
            # catch all, this probably isn't right but will not throw an
            # error
            return "S"

    def _serialize_obj(self, metadata, value):
        obj = ET.Element("Obj", RefId=self._get_obj_id())

        if len(value._types) > 0:
            self._create_tn(obj, value._types)

        to_string_value = value._to_string
        if to_string_value is not None:
            ET.SubElement(obj, "ToString").text = \
                self._serialize_string(to_string_value)

        for attr, property_meta in value._property_sets:
            attr_value = getattr(value, attr)
            self._create_obj(obj, attr_value, meta=property_meta)

        def serialize_prop(parent, properties):
            if len(properties) == 0:
                return
            parent = ET.SubElement(obj, parent)
            for attr, property_meta in properties:
                attr_value = getattr(value, attr)
                self._create_obj(parent, attr_value, meta=property_meta)

        serialize_prop("MS", value._extended_properties)
        serialize_prop("Props", value._adapted_properties)

        return obj

    def _serialize_dynamic_obj(self, metadata, value):
        obj = ET.Element("Obj", RefId=self._get_obj_id())
        self.obj[obj.attrib["RefId"]] = value

        if len(value.types) > 0:
            self._create_tn(obj, value.types)

        if value.to_string is not None:
            ET.SubElement(obj, "ToString").text = \
                self._serialize_string(value.to_string)

        for prop in value.property_sets:
            self._create_obj(obj, prop)

        def set_properties(element, prop_name):
            prop_keys = list(getattr(value, prop_name).keys())
            if len(prop_keys) == 0:
                return

            parent = ET.SubElement(obj, element)
            prop_keys.sort()
            for key in prop_keys:
                prop = getattr(value, prop_name)[key]
                self._create_obj(parent, prop, key=key)

        set_properties("MS", "extended_properties")
        set_properties("Props", "adapted_properties")

        return obj

    def _serialize_que(self, metadata, values):
        obj = ET.Element("Obj", RefId=self._get_obj_id())
        if not isinstance(metadata, QueueMeta):
            metadata = QueueMeta(name=metadata.name,
                                 optional=metadata.optional)
        self._create_tn(obj, metadata.list_types)

        que = ET.SubElement(obj, "QUE")
        while True:
            try:
                value = values.get(block=False)
                self.serialize(value, metadata.list_value_meta, parent=que,
                               clear=False)
            except Empty:
                break

        return obj

    def _serialize_stk(self, metadata, values):
        obj = ET.Element("Obj", RefId=self._get_obj_id())
        self._create_tn(obj, metadata.list_types)

        stk = ET.SubElement(obj, "STK")
        while True:
            try:
                value = values.pop()
                self.serialize(value, metadata.list_value_meta, parent=stk,
                               clear=False)
            except IndexError:
                break

        return obj

    def _serialize_ie(self, metadata, values):
        return self._serialize_lst(metadata, values, tag="IE")

    def _serialize_lst(self, metadata, values, tag="LST"):
        obj = ET.Element("Obj", RefId=self._get_obj_id())
        if not isinstance(metadata, ListMeta):
            metadata = ListMeta(name=metadata.name,
                                optional=metadata.optional)
        self._create_tn(obj, metadata.list_types)

        lst = ET.SubElement(obj, tag)
        for value in iter(values):
            entry_meta = copy(metadata.list_value_meta)
            self.serialize(value, entry_meta, parent=lst,
                           clear=False)

        return obj

    def _serialize_dct(self, metadata, values):
        obj = ET.Element("Obj", RefId=self._get_obj_id())
        if not isinstance(metadata, DictionaryMeta):
            metadata = DictionaryMeta(name=metadata.name,
                                      optional=metadata.optional)
        self._create_tn(obj, metadata.dict_types)

        dct = ET.SubElement(obj, "DCT")

        # allow dicts to be defined as a tuple so that the order is kept
        if isinstance(values, tuple):
            iterator = values
        else:
            iterator = values.items()

        for key, value in iterator:
            en = ET.SubElement(dct, "En")
            key_meta = copy(metadata.dict_key_meta)
            value_meta = copy(metadata.dict_value_meta)
            self.serialize(key, key_meta, parent=en, clear=False)
            self.serialize(value, value_meta, parent=en, clear=False)

        return obj

    def _serialize_string(self, value):
        if value is None:
            return None

        def rplcr(matchobj):
            surrogate_char = matchobj.group(0)
            byte_char = to_bytes(surrogate_char, encoding='utf-16-be')
            hex_char = to_unicode(binascii.hexlify(byte_char)).upper()
            hex_split = [hex_char[i:i + 4] for i in range(0, len(hex_char), 4)]

            return u"".join([u"_x%s_" % i for i in hex_split])

        # before running the translation we need to make sure _ before x is
        # encoded, normally _ isn't encoded except when preceding x
        string_value = to_unicode(value)

        # The MS-PSRP docs don't state this but the _x0000_ matcher is case insensitive so we need to make sure we
        # escape _X as well as _x.
        string_value = re.sub(u"(?i)_(x)", u"_x005F_\\1", string_value)
        string_value = re.sub(self._serial_str, rplcr, string_value)

        return string_value

    def _serialize_secure_string(self, value):
        if self.cipher is None:
            raise SerializationError("Cannot generate secure string as cipher "
                                     "is not initialised")

        # convert the string to a UTF-16 byte string as that is what is
        # expected in Windows. If a byte string (native string in Python 2) was
        # passed in, the sender must make sure it is a valid UTF-16
        # representation and not UTF-8 or else the server will fail to decrypt
        # the secure string in most cases
        string_bytes = to_bytes(value, encoding='utf-16-le')

        padder = PKCS7(self.cipher.algorithm.block_size).padder()
        padded_data = padder.update(string_bytes) + padder.finalize()

        encryptor = self.cipher.encryptor()
        ss_value = encryptor.update(padded_data) + encryptor.finalize()
        ss_string = to_string(base64.b64encode(ss_value))

        return ss_string

    def _deserialize_obj(self, element, metadata):
        obj = metadata.object()
        self.obj[element.attrib['RefId']] = obj

        to_string_value = element.find("ToString")
        if to_string_value is not None:
            obj._to_string = self._deserialize_string(to_string_value.text)

        def deserialize_property(prop_tag, properties):
            for attr, property_meta in properties:
                if attr == "invocation_info":
                    a = ""
                property_name = "Unknown"
                property_filter = ""
                if property_meta.name is not None:
                    property_name = property_meta.name
                    property_filter = "[@N='%s']" % property_meta.name

                tags = [property_meta.tag]
                # The below tags are actually seen as Obj in the parent element
                if property_meta.tag in ["DCT", "LST", "IE", "QUE", "STK",
                                         "ObjDynamic"]:
                    tags = ["Obj", "Ref"]

                val = None
                for tag in tags:
                    val = element.find("%s%s%s" % (prop_tag, tag,
                                                   property_filter))
                    if val is not None:
                        break

                if val is None and not property_meta.optional:
                    val = element.find("%sNil%s" % (prop_tag, property_filter))
                    if val is None:
                        obj_name = str(obj) if obj._to_string is not None \
                            else "Unknown"
                        err_msg = "Mandatory return value for '%s' was not " \
                                  "found on object %s"\
                                  % (property_name, obj_name)
                        raise SerializationError(err_msg)
                    val = None
                elif val is not None:
                    val = self.deserialize(val, property_meta, clear=False)

                setattr(obj, attr, val)

        deserialize_property("", obj._property_sets)
        deserialize_property("Props/", obj._adapted_properties)
        deserialize_property("MS/", obj._extended_properties)

        return obj

    def _deserialize_dynamic_obj(self, element, metadata):
        obj = metadata.object()
        self.obj[element.attrib['RefId']] = obj

        for obj_property in element:
            if obj_property.tag == "TN":
                for obj_type in obj_property:
                    obj.types.append(obj_type.text)
                self.tn[obj_property.attrib['RefId']] = obj.types
            elif obj_property.tag == "TNRef":
                obj.types = self.tn[obj_property.attrib['RefId']]
            elif obj_property.tag == "Props":
                for adapted_property in obj_property:
                    key = adapted_property.attrib['N']
                    value = self.deserialize(adapted_property, clear=False)
                    obj.adapted_properties[key] = value
            elif obj_property.tag == "MS":
                for extended_property in obj_property:
                    key = extended_property.attrib['N']
                    value = self.deserialize(extended_property, clear=False)
                    obj.extended_properties[key] = value
            elif obj_property.tag == "ToString":
                value = self.deserialize(obj_property, clear=False)
                obj.to_string = value
            else:
                value = self.deserialize(obj_property, clear=False)
                obj.property_sets.append(value)

        return obj

    def _deserialize_lst(self, element, metadata=None):
        list_value = []
        value_meta = getattr(metadata, "list_value_meta", None)

        entries = element.find("LST")
        for entry in entries:
            entry_value = self.deserialize(entry, value_meta, clear=False)
            list_value.append(entry_value)

        return list_value

    def _deserialize_que(self, element):
        queue = Queue()

        entries = element.find("QUE")
        for entry in entries:
            entry_value = self.deserialize(entry, clear=False)
            queue.put(entry_value)

        return queue

    def _deserialize_stk(self, element):
        # no native Stack object in Python so just use a list
        stack = []

        entries = element.find("STK")
        for entry in entries:
            entry_value = self.deserialize(entry, clear=False)
            stack.append(entry_value)

        return stack

    def _deserialize_dct(self, element):
        dictionary = {}
        entries = element.findall("DCT/En")
        for entry in entries:
            key = entry.find("*[@N='Key']")
            value = entry.find("*[@N='Value']")

            key = self.deserialize(key, clear=False)
            value = self.deserialize(value, clear=False)
            dictionary[key] = value

        return dictionary

    def _deserialize_string(self, value):
        if value is None:
            return ""

        def rplcr(matchobj):
            # The matched object is the UTF-16 byte representation of the UTF-8
            # hex string value. We need to decode the byte str to unicode and
            # then unhexlify that hex string to get the actual bytes of the
            # _x****_ value, e.g.
            # group(0) == b"\x00_\x00x\x000\x000\x000\x00A\x00_"
            # group(1) == b"\x000\x000\x000\x00A"
            # unicode (from utf-16-be) == u"000A"
            # returns b"\x00\x0A"
            match_hex = matchobj.group(1)
            hex_string = to_unicode(match_hex, encoding='utf-16-be')
            return binascii.unhexlify(hex_string)

        # need to ensure we start with a unicode representation of the string
        # so that we can get the actual UTF-16 bytes value from that string
        unicode_value = to_unicode(value)
        unicode_bytes = to_bytes(unicode_value, encoding='utf-16-be')
        bytes_value = re.sub(self._deserial_str, rplcr, unicode_bytes)
        return to_unicode(bytes_value, encoding='utf-16-be')

    def _deserialize_secure_string(self, value):
        if self.cipher is None:
            # cipher is not set up so we can't decrypt the string, just return
            # the raw element
            return value

        ss_string = base64.b64decode(value.text)
        decryptor = self.cipher.decryptor()
        decrypted_bytes = decryptor.update(ss_string) + decryptor.finalize()

        unpadder = PKCS7(self.cipher.algorithm.block_size).unpadder()
        unpadded_bytes = unpadder.update(decrypted_bytes) + unpadder.finalize()
        decrypted_string = to_unicode(unpadded_bytes, 'utf-16-le')

        return decrypted_string

    def _clear(self):
        self.obj_id = 0
        self.obj = {}
        self.tn = {}
        self.tn_id = 0

    def _get_obj_id(self):
        ref_id = str(self.obj_id)
        self.obj_id += 1
        return ref_id

    def _get_types_from_obj(self, element):
        obj_types = [e.text for e in element.findall("TN/T")]

        if len(obj_types) > 0:
            ref_id = element.find("TN").attrib['RefId']
            self.tn[ref_id] = obj_types

        tn_ref = element.find("TNRef")
        if tn_ref is not None:
            ref_id = tn_ref.attrib['RefId']
            obj_types = self.tn[ref_id]

        return obj_types

    def _create_tn(self, parent, types):
        main_type = types[0]
        ref_id = self.tn.get(main_type, None)
        if ref_id is None:
            ref_id = self.tn_id
            self.tn_id += 1
            self.tn[main_type] = ref_id

            tn = ET.SubElement(parent, "TN", RefId=str(ref_id))
            for type_name in types:
                ET.SubElement(tn, "T").text = type_name
        else:
            ET.SubElement(parent, "TNRef", RefId=str(ref_id))

    def _create_obj(self, parent, obj, key=None, meta=None):
        if isinstance(obj, ComplexObject):
            for ref, value in self.obj.items():
                if value == obj:
                    sub_element = ET.SubElement(parent, "Ref", RefId=ref)
                    if key is not None:
                        sub_element.attrib["N"] = key
                    return

        if meta is None:
            meta = ObjectMeta(name=key)
        self.serialize(obj, metadata=meta, parent=parent, clear=False)



# Copyright: (c) 2018, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import logging
import uuid

from pypsrp.complex_objects import Color, Coordinates, ObjectMeta, Size

log = logging.getLogger(__name__)


class PSHost(object):

    def __init__(self, current_culture, current_ui_culture, debugger_enabled,
                 name, private_data, ui, version):
        """
        Defines the properties and facilities provided by an application
        hosting a RunspacePool.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost

        This is a basic implementation some methods being noop or not
        implemented.

        :param current_culture: pypsrp.complex_objects.CultureInfo, the host's
            culture
        :param current_ui_culture: pypsrp.complex_objects.CultureInfo, the
            host's UI culture
        :param debugger_enabled: This property enables and disables the host
            debugger if debugging is supported
        :param name: Gets the hosting application identification in some user-
            friendly fashion.
        :param private_data: Used to allow the host to pass private data
            through a Runspace to cmdlets inside that Runspace
        :param ui: The hosts implementation of PSHostUserInterface. Should be
            None if the host that does not want to support user interaction
        :param version: The version of the hosting application
        """
        self.ui = ui
        self.debugger_enabled = debugger_enabled
        self.private_data = private_data
        self.rc = None

        self.name = name
        self.version = version
        self.instance_id = uuid.uuid4()
        self.current_culture = current_culture
        self.current_ui_culture = current_ui_culture

    def run_method(self, method_identifier, args, runspace, pipeline=None):
        """
        Run a host call method requested by the server and return the response
        from this method to send back to the server.
        https://msdn.microsoft.com/en-us/library/dd306624.aspx

        Each method will have access to the current runspace and pipeline (if
        applicable) during the method call as well as any args sent from the
        server.

        :param method_identifier: pypsrp.complex_objects.HostMethodIdentifier
            in the host call message.
        :param args: The list of arguments for the host call function.
        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :return: The response (if any) to send back to the server
        """
        response = None
        if method_identifier.value < 11:
            func = getattr(self, str(method_identifier))
            response = func(runspace, pipeline, *args)
        elif method_identifier.value < 27:
            func = getattr(self.ui, str(method_identifier))
            response = func(runspace, pipeline, *args)
        elif method_identifier.value < 52:
            func = getattr(self.ui.raw_ui, str(method_identifier))
            response = func(runspace, pipeline, *args)
        else:
            log.warning("Received unexpected/unsupported host method "
                        "identifier: %d" % method_identifier.value)

        return response

    # Start of Host Methods, the names of these functions are important as
    # they line up to the names defined by MS and are sent in the host call
    # messages
    def GetName(self, runspace, pipeline):
        """
        MI: 1
        SHOULD return a string identifying the hosting application in a user
        friendly way.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost.name

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :return: String of the user-friendly name of the hosting application
        """
        return self.name

    def GetVersion(self, runspace, pipeline):
        """
        MI: 2
        SHOULD return the version number of the hosting application.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost.version

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :return: Version number of the hosting application
        """
        meta = ObjectMeta("Version")
        value = runspace.serialize(self.version, meta)
        return value

    def GetInstanceId(self, runspace, pipeline):
        """
        MI: 3
        SHOULD return a GUID that uniquely identifies the hosting application.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost.instanceid

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :return: GUID of the hosting application
        """
        return self.instance_id

    def GetCurrentCulture(self, runspace, pipeline):
        """
        MI: 4
        SHOULD return the host's culture.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost.currentculture

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :return: pypsrp.complex_objects.CultureInfo of the host's culture
        """
        return self.current_culture

    def GetCurrentUICulture(self, runspace, pipeline):
        """
        MI: 5
        MUST return the host's UI culture.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost.currentuiculture

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :return: pypsrp.complex_objects.CultureInfo of the host's UI culture
        """
        return self.current_ui_culture

    def SetShouldExit(self, runspace, pipeline, exit_code):
        """
        MI: 6
        SHOULD shut down the hosting application and close the current
        runspace. The default implementation just sets the rc on the host
        object and doesn't shutdown the runspace.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost.setshouldexit

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param exit_code: The exit code accompanying the exit keyword.
            Typically after exiting a runspace, a host will also terminate
        """
        self.rc = exit_code

    def EnterNestedPrompt(self, runspace, pipeline):
        """
        MI: 7
        SHOULD interrupt the current pipeline and start a nested pipeline.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost.enternestedprompt

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        """
        raise NotImplementedError()

    def ExitNestedPrompt(self, runspace, pipeline):
        """
        MI: 8
        SHOULD stop the nested pipeline and resume the current pipeline.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost.exitnestedprompt

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        """
        raise NotImplementedError()

    def NotifyBeginApplication(self, runspace, pipeline):
        """
        MI: 9
        Called by an application to indicate that it is executing a command
        line application.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost.notifybeginapplication

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        """
        pass

    def NotifyEndApplication(self, runspace, pipeline):
        """
        MI: 10
        Called by an application to indicate that it has finished executing a
        command line application.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost.notifyendapplication

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        """
        pass


class PSHostUserInterface(object):

    def __init__(self, raw_ui=None):
        """
        Defines the properties and facilities provided by a hosting application
        deriving from PSHost that offers dialog-oriented and line-oriented
        interactive features.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface

        This is a basic implementation some methods being noop or not
        implemented.

        :param raw_ui: Implementation of PSHostRawUserInterface, set to None
            if there is no raw user interface
        """
        self.raw_ui = raw_ui

        # the below properties don't need to be used, they are just here for
        # the default implementation
        self.stdout = []
        self.stderr = []

    def ReadLine(self, runspace, pipeline):
        """
        MI: 11
        SHOULD read a line of characters from a user.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.readline

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :return: A string of characters to return to the read line call
        """
        raise NotImplementedError()

    def ReadLineAsSecureString(self, runspace, pipeline):
        """
        MI: 12
        SHOULD read a line of characters from a user, with the user input not
        echoed.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.readlineassecurestring

        Because the return value is meant to be a SecureString, the user must
        either have called or will call runspace.exchange_keys() in this
        implementation so that the serializer can create the string.

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :return: The characters types by the user in an encrypted form
        """
        raise NotImplementedError()

    def Write1(self, runspace, pipeline, value):
        """
        MI: 13
        SHOULD write specified characters on the hosting application.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.write

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param value: The string of characters to be written
        """
        self.stdout.append(value)

    def Write2(self, runspace, pipeline, foreground_color, background_color,
               value):
        """
        MI: 14
        SHOULD write the specified characters with the specified foreground and
        background color on the hosting application.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.write

        This implementation just adds this result to the stdout list and
        ignores the colors, create your own method implementation if you wish
        to utilise this correctly

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param foreground_color: The int value of pypsrp.complex_objects.Color
            of the foreground color to display the text with
        :param background_color: The int value of pypsrp.complex_objects.Color
            of the background color to display the text with
        :param value: The string of characters to be written
        """
        self.stdout.append(value)

    def WriteLine1(self, runspace, pipeline):
        """
        MI: 15
        SHOULD write a carriage return on the hosting application.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.writeline

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        """
        self.stdout.append("\r\n")

    def WriteLine2(self, runspace, pipeline, value):
        """
        MI: 16
        SHOULD write the specified line on the hosting application.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.writeline

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param value: The string of characters to be written
        """
        self.stdout.append(value + "\r\n")

    def WriteLine3(self, runspace, pipeline, foreground_color,
                   background_color, value):
        """
        MI: 17
        SHOULD write the specified line with the specified foreground and
        background color on the hosting application.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.writeline

        This implementation just adds this result to the stdout list and
        ignores the colors, create your own method implementation if you wish
        to utilise this correctly

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param foreground_color: The int value of pypsrp.complex_objects.Color
            of the foreground color to display the text with
        :param background_color: The int value of pypsrp.complex_objects.Color
            of the background color to display the text with
        :param value: The string of characters to be written
        """
        self.stdout.append(value + "\r\n")

    def WriteErrorLine(self, runspace, pipeline, message):
        """
        MI: 18
        SHOULD write a line to the error display of the hosting application.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.writeerrorline

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param message: The message to display
        """
        self.stderr.append(message + "\r\n")

    def WriteDebugLine(self, runspace, pipeline, message):
        """
        MI: 19
        SHOULD write a line to the debug display of the hosting application.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.writedebugline

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param message: The message to display
        """
        self.stdout.append("DEBUG: %s\r\n" % message)

    def WriteProgress(self, runspace, pipeline, source_id, record):
        """
        MI: 20
        SHOULD display a progress record on the hosting application.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.writeprogress

        Because of the way MS serializes the record in the method call args,
        the value for record is the serialized XML string for a ProgressRecord
        message. You can manually parse it like;

            from pypsrp.messages import ProgressRecord

            meta = ObjectMeta("Obj", object=ProgressRecord)
            rec = runspace._serializer.deserialize(record, meta)

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param source_id: Unique identifier of the source of the record
        :param record: A ProgressRecord serialized as XML
        """
        pass

    def WriteVerboseLine(self, runspace, pipeline, message):
        """
        MI: 21
        SHOULD write a line on the verbose display of the hosting application.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.writeverboseline

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param message: The verbose message to display
        """
        self.stdout.append("VERBOSE: %s\r\n" % message)

    def WriteWarningLine(self, runspace, pipeline, message):
        """
        MI: 22
        SHOULD write a line on the warning display of the hosting application.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.writewarningline

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param message: The warning message to display
        """
        self.stdout.append("WARNING: %s\r\n" % message)

    def Prompt(self, runspace, pipeline, caption, message, descriptions):
        """
        MI: 23
        SHOULD prompt the user with a set of choices.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.prompt

        The descriptions arg is a list of GenericComplexObjects with the
        following extended attributes (correlates to FieldDescription in .NET):
            attributes
            defaultValue
            helpMessage
            isMandatory
            label
            name
            parameterAssemblyFullName
            parameterTypeFullName
            parameterTypeName

        For example you can access the prompt name from `Read-Host -Prompt`
        with descriptions[i].extended_properties['name'].

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param caption: Caption to precede or title the prompt
        :param message: A text description of the set of fields to be prompted
        :param descriptions: list of serialized FieldDescriptions that contain
            information about each field to be prompted for
        :return: Dict with results of prompting. Key are the field names from
            the FieldDescriptions, the values are objects representing the
            values of the corresponding fields as collected from the user.
        """
        raise NotImplementedError()

    def PromptForCredential1(self, runspace, pipeline, caption, message,
                             user_name, target_name):
        """
        MI: 24
        SHOULD prompt the user for entering credentials with the specified
        caption, message, user name and target name.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.promptforcredential

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param caption: Caption for the message
        :param message: Text description for the credential to be prompted
        :param user_name: Name of the user whose credential is to be prompted
            for. If set to null or empty string, the function will prompt for
            the user name first
        :param target_name: Name of the target for which the credential is
            being collected
        :return: PSCredential object of the user input credential
        """
        raise NotImplementedError()

    def PromptForCredential2(self, runspace, pipeline, caption, message,
                             user_name, target_name, allowed_credential_types,
                             options):
        """
        MI: 25
        SHOULD prompt the user for entering credentials with the specified
        caption, message, username, target name, allowed credential types and
        options.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.promptforcredential

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param caption: Caption for the message
        :param message: Text description for the credential to be prompted
        :param user_name: Name of the user whose credential is to be prompted
            for. If set to null or empty string, the function will prompt for
            the user name first
        :param target_name: Name of the target for which the credential is
            being collected
        :param allowed_credential_types: the int value for PSCredentialTypes,
            types of credentials that can be supplied by the user
        :param options: the int value for PSCredentialUIOptions, options that
            control the credential gathering UI behavior
        :return: PSCredential object of the user input credential
        """
        raise NotImplementedError()

    def PromptForChoice(self, runspace, pipeline, caption, message, choices,
                        default_choice):
        """
        MI: 26
        SHOULD display a list of choices to the user and MUST return the index
        of the selected option.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.promptforchoice

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param caption: The caption to precede or title the prompt
        :param message: A message that describes what the choice is for
        :param choices: A list of serialized (GenericComplexObject)
            ChoiceDescription objects that describe each choice
        :param default_choice: The index of the label in the choices collection
            element to be present to the user as the default choice, -1 means
            no default
        :return: The index of the choices element that corresponds to the
            option selected
        """
        raise NotImplementedError()


class PSHostRawUserInterface(object):

    def __init__(self, window_title, cursor_size, foreground_color,
                 background_color, cursor_position, window_position,
                 buffer_size, max_physical_window_size, max_window_size,
                 window_size):
        """
        Defines the lowest-level user interface functions that an interactive
        application hosting a Runspace can choose to implement if it wants
        to support any cmdlet that does character-mode interaction with the
        user.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface

        This is a basic framework implementation with the majority of the
        methods not implemented or tested.

        :param window_title: The titlebar text of the current view window
        :param cursor_size: The size of the cursor as a percentage (0 to 100)
        :param foreground_color: The pypsrp.complex_objects.Color used to
            render characters on the screen buffer
        :param background_color: The pypsrp.complex_objects.Color used to
            render the backfround behind characters on the screen buffer
        :param cursor_position: The pypsrp.complex_objects.Coordinates of the
            position of the cursor in the screen buffer
        :param window_position: The pypsrp.complex_objects.Coordinates of the
            position of the window relative to the screen buffer, (0, 0) is the
            upper left of the screen buffer
        :param buffer_size: The pypsrp.complex_objects.Size of the screen
            buffer
        :param max_physical_window_size: The pypsrp.complex_objects.Size of the
            largest windows possible for the display hardward
        :param max_window_size: The pypsrp.complex_objects.Size of the window
            possible for the current buffer
        :param window_size: The pypsrp.complex_objects.Size of the current
            window, cannot be larger than max_physical_window_size
        """
        self.key_available = False

        self.window_title = window_title
        self.cursor_size = cursor_size
        self.foreground_color = foreground_color
        self.background_color = background_color
        self.cursor_position = cursor_position
        self.window_position = window_position
        self.buffer_size = buffer_size
        self.max_physical_window_size = max_physical_window_size
        self.max_window_size = max_window_size
        self.window_size = window_size

    def GetForegroundColor(self, runspace, pipeline):
        """
        MI: 27
        SHOULD return the foreground color of the hosting application.

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :return: A pypsrp.complex_objects.Color to return to the server
        """
        return self.foreground_color

    def SetForegroundColor(self, runspace, pipeline, color):
        """
        MI: 28
        SHOULD set the foreground color of the hosting application.

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param color: The int value for pypsrp.complex_objects.Color to set
        """
        self.foreground_color = Color(value=color)

    def GetBackgroundColor(self, runspace, pipeline):
        """
        MI: 29
        SHOULD return the background color of the hosting application.

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :return: A pypsrp.complex_objects.Color to return to the server
        """
        return self.background_color

    def SetBackgroundColor(self, runspace, pipeline, color):
        """
        MI: 30
        SHOULD set the background color of the hosting application.

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param color: The int value for pypsrp.complex_objects.Color to set
        """
        self.background_color = Color(value=color)

    def GetCursorPosition(self, runspace, pipeline):
        """
        MI: 31
        SHOULD return the current cursor position in the hosting application.

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :return:  A pypsrp.complex_objects.Coordinates to return to the server
        """
        return self.cursor_position

    def SetCursorPosition(self, runspace, pipeline, coordinates):
        """
        MI: 32
        SHOULD return the current cursor position in the hosting application.

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param coordinates: A GenericComplexObject that contains the extended
            properties for the coordinates
        """
        pos = Coordinates(x=coordinates.extended_properties['x'],
                          y=coordinates.extended_properties['y'])
        self.cursor_position = pos

    def GetWindowPosition(self, runspace, pipeline):
        """
        MI: 33
        SHOULD return the position of the view window relative to the screen
        buffer.

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :return: A pypsrp.complex_objects.Coordinates to return to the server
        """
        return self.window_position

    def SetWindowPosition(self, runspace, pipeline, coordinates):
        """
        MI: 34
        SHOULD set the position of the view window relative to the screen
        buffer.

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param coordinates: A GenericComplexObject that contains the extended
            properties for the coordinates
        """
        pos = Coordinates(x=coordinates.extended_properties['x'],
                          y=coordinates.extended_properties['y'])
        self.window_position = pos

    def GetCursorSize(self, runspace, pipeline):
        """
        MI: 35
        SHOULD return the cursor size as a percentage.

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :return: The int value for the cursor size
        """
        return self.cursor_size

    def SetCursorSize(self, runspace, pipeline, percentage):
        """
        MI: 36
        SHOULD set the cursor size based on the percentage value specified.

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param percentage: The int value representing the cursor size
        """
        self.cursor_size = percentage

    def GetBufferSize(self, runspace, pipeline):
        """
        MI: 37
        SHOULD return the current size of the screen buffer, measured in
        character cells.

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :return: pypsrp.complex_object.Size of the screen buffer, measured in
            character cells
        """
        return self.buffer_size

    def SetBufferSize(self, runspace, pipeline, size):
        """
        MI: 38
        SHOULD set the size of the screen buffer with the specified size in
        character cells.

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param size: A GenericComplexObject that contains the extended
            properties for the size
        """
        obj = Size(height=size.extended_properties['height'],
                   width=size.extended_properties['width'])
        self.buffer_size = obj

    def GetWindowSize(self, runspace, pipeline):
        """
        MI: 39
        SHOULD return the current view window size.

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :return: pypsrp.complex_objects.Size of the current window
        """
        return self.window_size

    def SetWindowSize(self, runspace, pipeline, size):
        """
        MI: 40
        SHOULD set the view window size based on the size specified.

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param size: A GenericComplexObject that contains the extended
            properties for the size
        """
        obj = Size(height=size.extended_properties['height'],
                   width=size.extended_properties['width'])
        self.window_size = obj

    def GetWindowTitle(self, runspace, pipeline):
        """
        MI: 41
        SHOULD return the title of the hosting application's window.

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :return: The window title of the hosting application
        """
        return self.window_title

    def SetWindowTitle(self, runspace, pipeline, title):
        """
        MI: 42
        SHOULD set the view window size based on the size specified.

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param title: The string for the window title to set
        """
        self.window_title = title

    def GetMaxWindowSize(self, runspace, pipeline):
        """
        MI: 43
        SHOULD return the maximum window size possible for the current buffer,
        current font, and current display hardware.

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :return: pypsrp.complex_objects.Size of the max possible window size
        """
        return self.max_window_size

    def GetMaxPhysicalWindowSize(self, runspace, pipeline):
        """
        MI: 44
        SHOULD return the maximum window size possible for the current font and
        current display hardware, ignoring the current buffer size.

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :return: pypsrp.complex_objects.Size of the max physically possible
            window size
        """
        return self.max_physical_window_size

    def GetKeyAvailable(self, runspace, pipeline):
        """
        MI: 45
        SHOULD examine if a keystroke is waiting on the input, returning TRUE
        if so and FALSE otherwise

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :return: bool if a keystroke is waiting on the input
        """
        return self.key_available

    def ReadKey(self, runspace, pipeline, options=4):
        """
        MI: 46
        SHOULD read a key stroke from the keyboard, blocking until a key is
        typed.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.readkey

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param options: Bit mask combo of ReadKeyOptions, default is
            ReadKeyOptions.IncludeKeyDown
        :return: KeyInfo - key stroke depending on the value of options
        """
        raise NotImplementedError()

    def FlushInputBuffer(self, runspace, pipeline):
        """
        MI: 47
        SHOULD reset the keyboard input buffer.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.flushinputbuffer

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        """
        pass

    def SetBufferContents1(self, runspace, pipeline, rectangle, fill):
        """
        MI: 49
        SHOULD copy the specified buffer cell into all the cells within the
        specified rectangle.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.setbuffercontents

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param rectange: A GenericComplexObject that represents the rectangle
            in which to fill. If all element are -1, the entire screen buffer
            will be copied with fill. Contains the following extended
            properties: left, top, right, bottom
        :param fill: A GenericComplexObject of the characters and attributes
            used to fill the rectangle
        """
        pass

    def SetBufferContents2(self, runspace, pipeline, origin, contents):
        """
        MI: 48
        SHOULD copy the specified buffer cell array into the screen buffer at
        the specified coordinates (as specified in section 2.2.3.1).
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.setbuffercontents

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param origin: A GenericComplexObject that represents the coordinates,
            x and y are extended properties of this object
                origin.extended_properties['x']
                origin.extended_properties['y']
        :param contents: A rectangle of BufferCell objects to be copied to the
            screen buffer. This is also a GenericComplexObject which is a multi
            dimensional array.
            https://msdn.microsoft.com/en-us/library/dd340684.aspx
                # number of elements in each row
                contents.extended_properties['mal']

                # each BufferCell is in this list, use mal to determine what
                # rows they are in
                contents.extended_properties['mae']
        """
        pass

    def GetBufferContents(self, runspace, pipeline, rectangle):
        """
        MI: 50
        SHOULD return the contents in a specified rectangular region of the
        hosting application's window and MUST return an array of buffer cells.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.getbuffercontents

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param rectangle: The rectangle on the screen buffer to extract
        :return: A pypsrp.complex_objects.Array of BufferCell objects extracted
            from the rectangular region of the screen buffer specified by
            rectangle
        """
        raise NotImplementedError()

    def ScrollBufferContents(self, runspace, pipeline, source, destination,
                             clip, fill):
        """
        MI: 51
        SHOULD scroll a region on the screen buffer.
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.scrollbuffercontents

        :param runspace: The runspace the host call relates to
        :param pipeline: The pipeline (if any) that the call relates to
        :param source: Rectangle - Indicates the region of the screen to be
            scrolled
        :param destination: Coordinates - Indicates the upper left coordinates
            of the region of the screen to receive the source region contents.
            That target region is the same size as the source region
        :param clip: Rectangle - Indicates the region of the screen to include
            in the operation. If a cell would be changed by the operation but
            does not fall within the clip region, it will be unchanged
        :param fill: BufferCell - The character and attributes to be used to
            fill any cells within the intersection of the source rectangle and
            clipping rectangle that are left "empty" by the move
        """
        pass



# Copyright: (c) 2018, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import base64
import logging
import sys

from pypsrp.exceptions import WSManFaultError
from pypsrp.wsman import NAMESPACES, OptionSet, SelectorSet

if sys.version_info[0] == 2 and sys.version_info[1] < 7:  # pragma: no cover
    # ElementTree in Python 2.6 does not support namespaces so we need to use
    # lxml instead for this version
    from lxml import etree as ET
else:  # pragma: no cover
    import xml.etree.ElementTree as ET

log = logging.getLogger(__name__)


class CommandState(object):
    DONE = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/" \
           "CommandState/Done"
    PENDING = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/" \
              "CommandState/Pending"
    RUNNING = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/" \
              "CommandState/Running"


class SignalCode(object):
    """
    [MS-WSMV] 2.2.4.38 Signal - Code
    https://msdn.microsoft.com/en-us/library/cc251558.aspx

    The control code to send in a Signal message to the server
    """
    CTRL_C = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/" \
             "signal/ctrl_c"
    CTRL_BREAK = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/" \
                 "signal/ctrl_break"
    TERMINATE = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/" \
                "signal/Terminate"
    PS_CTRL_C = "powershell/signal/ctrl_c"


class WinRS(object):

    def __init__(self, wsman, resource_uri="http://schemas.microsoft.com/wbem/"
                                           "wsman/1/windows/shell/cmd",
                 id=None, input_streams='stdin',
                 output_streams='stdout stderr', codepage=None,
                 environment=None, idle_time_out=None, lifetime=None,
                 name=None, no_profile=None, working_directory=None):
        """
        A WinRS shell instance. This is used by Process to spawn a new command/
        process on the raw WinRS shell.

        :param wsman: The pypsrp.wsman.WSMan instance to send commands over
        :param resource_uri: The resource URI of the shell, defaults to the
            WinRS cmd shell
        :param id: The ID of the shell, if not specified a dynamic ID will be
            generated by the host
        :param input_streams: The input streams available to the shell
        :param output_streams: The output streams available to the shell
        :param codepage: The codepage of the shell
        :param environment: A dictionary that contains environment key/values
            that are created for the shell instance
        :param idle_time_out: The idle timeout in seconds of the shell
        :param lifetime: The total lifetime of the shell
        :param name: The name (description) of the shell
        :param no_profile: Whether to create the shell with the user profile
            active or not
        :param working_directory: The default working directory of the created
            shell
        """
        self.wsman = wsman
        self.opened = False
        self.id = id
        self.resource_uri = resource_uri
        self.input_streams = input_streams
        self.output_streams = output_streams
        self.codepage = codepage
        self.environment = environment
        self.idle_time_out = idle_time_out
        self.lifetime = lifetime
        self.name = name
        self.no_profile = no_profile
        self.working_directory = working_directory
        self.owner = None
        self.client_ip = None
        self.shell_run_time = None
        self.shell_inactivity = None

        self._selector_set = None
        # TODO: should I store a process table like a RunspacePool

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        """
        Closes the shell
        """
        if not self.opened:
            return
        self.wsman.delete(self.resource_uri, selector_set=self._selector_set)
        self.id = None
        self.opened = False

    def command(self, executable, arguments, no_shell=False, command_id=None):
        """
        Send a command message to the Shell. Process should really be used
        instead if a normal WinRS process is desired.

        :param executable: The path to the command/executable
        :param arguments: A list of arguments to run with the executable
        :param no_shell: Whether to create the command in the cmd shell or
            bypass it. If True then executable must be the full path to the
            exe. This only works on older OS's before 2012 R2 (not including)
        :param command_id: The command ID to specify when creating the command
        :return: The raw WSMan body response
        """
        rsp = NAMESPACES['rsp']

        options = OptionSet()
        options.add_option('WINRS_SKIP_CMD_SHELL', no_shell)

        arguments = arguments if arguments is not None else []

        cmd = ET.Element("{%s}CommandLine" % rsp)
        if command_id is not None:
            cmd.attrib['CommandId'] = command_id

        ET.SubElement(cmd, "{%s}Command" % rsp).text = executable
        for argument in arguments:
            ET.SubElement(cmd, "{%s}Arguments" % rsp).text = argument

        return self.wsman.command(self.resource_uri, cmd, option_set=options,
                                  selector_set=self._selector_set)

    def open(self, base_options=None, open_content=None):
        """
        Send an open message to the WSMan host

        :param base_options: Any OptionSet options to pass along to the Open
            message
        :param open_content: Any extra XML elements to add to the Open message
        :return: The raw WSMan body response
        """
        if self.opened:
            return

        rsp = NAMESPACES['rsp']

        shell = ET.Element("{%s}Shell" % rsp)
        if self.id is not None:
            shell.attrib['ShellId'] = self.id

        ET.SubElement(shell, "{%s}InputStreams" % rsp).text = \
            self.input_streams
        ET.SubElement(shell, "{%s}OutputStreams" % rsp).text = \
            self.output_streams
        if self.environment is not None:
            env = ET.SubElement(shell, "{%s}Environment" % rsp)
            for key, value in self.environment.items():
                ET.SubElement(env, "{%s}Variable" % rsp,
                              Name=str(key)).text = str(value)

        if self.idle_time_out is not None:
            ET.SubElement(shell, "{%s}IdleTimeOut" % rsp).text = \
                "PT%sS" % str(self.idle_time_out)

        if self.lifetime is not None:
            ET.SubElement(shell, "{%s}Lifetime" % rsp).text = \
                "PT%sS" % self.lifetime

        if self.name is not None:
            ET.SubElement(shell, "{%s}Name" % rsp).text = self.name

        if self.working_directory is not None:
            ET.SubElement(shell, "{%s}WorkingDirectory" % rsp).text = \
                self.working_directory

        if open_content is not None:
            shell.append(open_content)

        # inherit the base options if it was passed in, otherwise use an empty
        # option set
        options = OptionSet() if base_options is None else base_options
        if self.no_profile is not None:
            options.add_option('WINRS_NOPROFILE', self.no_profile)
        if self.codepage is not None:
            options.add_option('WINRS_CODEPAGE', self.codepage)

        if len(options.values) == 0:
            # set options back to None if nothing was actually set
            options = None

        response = self.wsman.create(self.resource_uri, shell,
                                     option_set=options)
        self._parse_shell_create(response)
        self.opened = True

        return response

    def receive(self, stream='stdout stderr', command_id=None, timeout=None):
        """
        Send a receive message to the WSMan host

        :param stream: The stream(s) separated by a space to receive the
            response for
        :param command_id: If specified the COmmand ID to receive the response
            from the command and not the shell
        :param timeout: Override the default WSMan timeout on the receive
            command
        :return: A tuple of
            state: The command state on the received response
            rc: The return code (if any) on the received response
            buffer: A dict containing a byte string for each buffer, the stream
                name of each buffer is the key in this return value
        """
        rsp = NAMESPACES['rsp']

        receive = ET.Element("{%s}Receive" % rsp)
        stream_xml = ET.SubElement(receive,
                                   "{%s}DesiredStream" % rsp)
        stream_xml.text = stream
        if command_id is not None:
            stream_xml.attrib['CommandId'] = command_id

        options = OptionSet()
        options.add_option('WSMAN_CMDSHELL_OPTION_KEEPALIVE', True)

        response = self.wsman.receive(self.resource_uri, receive,
                                      option_set=options,
                                      selector_set=self._selector_set,
                                      timeout=timeout)

        command_state = response.find("rsp:ReceiveResponse/"
                                      "rsp:CommandState",
                                      namespaces=NAMESPACES)
        if command_state is not None:
            command_state = command_state.attrib['State']

        rc = response.find("rsp:ReceiveResponse/"
                           "rsp:CommandState/"
                           "rsp:ExitCode",
                           namespaces=NAMESPACES)
        if rc is not None:
            rc = int(rc.text)

        buffer = {}
        for stream_name in stream.split(" "):
            buffer[stream_name] = b""
        streams = response.findall("rsp:ReceiveResponse/"
                                   "rsp:Stream",
                                   namespaces=NAMESPACES)
        for stream in streams:
            if stream.text is None:
                continue

            stream_value = base64.b64decode(stream.text.encode('utf-8'))
            stream_name = stream.attrib['Name']
            buffer[stream_name] += stream_value

        return command_state, rc, buffer

    def send(self, stream, data, command_id=None, end=None):
        """
        Send the input data to the shell or command (if command_id is set)

        :param stream: The stream to send the data to
        :param data: The byte string of the data to send
        :param command_id: If the input data is for a command, then this
            specifies the Command ID it is for
        :param end: Whether this is the last input element for the command
        :return: The raw WSMan body of the response
        """
        rsp = NAMESPACES['rsp']

        send = ET.Element("{%s}Send" % rsp)
        stream = ET.SubElement(send, "{%s}Stream" % rsp, Name=stream)
        if end is not None:
            stream.attrib['End'] = str(end)
        if command_id is not None:
            stream.attrib['CommandId'] = command_id

        stream.text = base64.b64encode(data).decode('utf-8')
        return self.wsman.send(self.resource_uri, send,
                               selector_set=self._selector_set)

    def signal(self, code, command_id):
        """
        Send a signal to the command

        :param code: The SignalCode value to send
        :param command_id: The command id the signal is for
        :return: The raw WSMan body of the response
        """
        rsp = NAMESPACES['rsp']

        signal = ET.Element("{%s}Signal" % rsp,
                            attrib={"CommandId": command_id})
        ET.SubElement(signal, "{%s}Code" % rsp).text = code
        return self.wsman.signal(self.resource_uri, signal,
                                 selector_set=self._selector_set)

    def _parse_shell_create(self, response):
        fields = {
            "rsp:ShellId": "id",
            "rsp:ResourceUri": "resource_uri",
            "rsp:Owner": "owner",
            "rsp:ClientIP": "client_ip",
            "rsp:IdleTimeOut": "idle_time_out",
            "rsp:OutputStreams": "output_streams",
            "rsp:ShellRunTime": "shell_run_time",
            "rsp:ShellInactivity": "shell_inactivity"
        }

        for xml_element, shell_attr in fields.items():
            element = response.find("rsp:Shell/%s" % xml_element, NAMESPACES)
            if element is not None:
                setattr(self, shell_attr, element.text)

        selector_set = response.find("wst:ResourceCreated/"
                                     "wsa:ReferenceParameters/"
                                     "wsman:SelectorSet", NAMESPACES)
        if selector_set is not None:
            self._selector_set = SelectorSet()
            for selector in selector_set:
                self._selector_set.add_option(selector.attrib['Name'],
                                              selector.text)


class Process(object):

    def __init__(self, shell, executable, arguments=None, id=None,
                 no_shell=False):
        """
        A new process to run over a default WinRS shell.

        :param shell: The WinRS shell to run the process over
        :param executable: The execute/command to run
        :param arguments: A list of arguments to use with the executable or
            command
        :param id: The ID of the command, if not specified then this is
            dynamically created
        :param no_shell: Whether to create the command in the cmd shell or
            bypass it. If True then executable must be the full path to the
            exe. This only works on older OS's before 2012 R2 (not including)
        """
        log.debug("Creating WinRS process for '%s' with arguments '%s'"
                  % (executable, arguments))
        self.shell = shell
        self.id = id
        self.no_shell = no_shell
        self.executable = executable
        self.arguments = arguments
        self.state = CommandState.PENDING
        self.rc = None
        self.stdout = b""
        self.stderr = b""

    def begin_invoke(self):
        """
        Start the process in the background and return immediately. Call
        poll_invoke to get the latest output/status of the command and
        end_invoke to wait until the process is complete.
        """
        response = self.shell.command(self.executable, self.arguments,
                                      no_shell=self.no_shell)
        self.id = response.find("rsp:CommandResponse/rsp:CommandId",
                                namespaces=NAMESPACES).text

    def end_invoke(self):
        """
        Wait until the process is done
        """
        while not self.state == CommandState.DONE:
            self.poll_invoke()

    def invoke(self):
        """
        Start the process synchronously and wait for it to be completed.
        """
        self.begin_invoke()
        self.end_invoke()

    def poll_invoke(self, timeout=None):
        """
        Poll the running process to update the output buffer and the status.

        :param timeout: Override the default WSMan timeout when polling the
        process.
        """
        try:
            self.state, self.rc, buffer = self.shell.receive('stdout stderr',
                                                             self.id,
                                                             timeout=timeout)
        except WSManFaultError as exc:
            # if a command exceeds the OperationTimeout set, we will get
            # a WSManFaultError with the code 2150858793. We ignore this
            # as it just meant no output during that operation.
            if exc.code == 2150858793:
                pass
            else:
                raise exc
        else:
            self.stdout += buffer['stdout']
            self.stderr += buffer['stderr']

    def send(self, data, end=True):
        """
        Send data to the running process.

        :param data: The byte string to send
        :param end: Whether this is the last input to send
        :return:
        """
        self.shell.send("stdin", data, command_id=self.id, end=end)

    def signal(self, code):
        """
        Send a signal to the process.

        :param code: The SignalCode to send
        """
        self.shell.signal(code, self.id)



# Copyright: (c) 2018, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import logging
import struct
import sys
import uuid
import warnings

from six import binary_type

from pypsrp.complex_objects import ApartmentState, CommandType, \
    ComplexObject, DictionaryMeta, ErrorRecord, GenericComplexObject, \
    HostInfo, HostMethodIdentifier, InformationalRecord, ListMeta, \
    ObjectMeta, Pipeline, ProgressRecordType, PSThreadOptions, \
    RemoteStreamOptions
from pypsrp.exceptions import SerializationError
from pypsrp._utils import to_string


if sys.version_info[0] == 2 and sys.version_info[1] < 7:  # pragma: no cover
    # ElementTree in Python 2.6 does not support namespaces so we need to use
    # lxml instead for this version
    from lxml import etree as ET
else:  # pragma: no cover
    import xml.etree.ElementTree as ET

log = logging.getLogger(__name__)


class Destination(object):
    # The destination of a PSRP message
    CLIENT = 0x00000001
    SERVER = 0x00000002


class MessageType(object):
    """
    [MS-PSRP] 2.2.1 PowerShell Remoting Protocol Message - MessageType
    https://msdn.microsoft.com/en-us/library/dd303832.aspx

    Identifier of the message contained within a PSRP message
    """
    SESSION_CAPABILITY = 0x00010002
    INIT_RUNSPACEPOOL = 0x00010004
    PUBLIC_KEY = 0x00010005
    ENCRYPTED_SESSION_KEY = 0x00010006
    PUBLIC_KEY_REQUEST = 0x00010007
    CONNECT_RUNSPACEPOOL = 0x00010008
    RUNSPACEPOOL_INIT_DATA = 0x002100B
    RESET_RUNSPACE_STATE = 0x0002100C
    SET_MAX_RUNSPACES = 0x00021002
    SET_MIN_RUNSPACES = 0x00021003
    RUNSPACE_AVAILABILITY = 0x00021004
    RUNSPACEPOOL_STATE = 0x00021005
    CREATE_PIPELINE = 0x00021006
    GET_AVAILABLE_RUNSPACES = 0x00021007
    USER_EVENT = 0x00021008
    APPLICATION_PRIVATE_DATA = 0x00021009
    GET_COMMAND_METADATA = 0x0002100A
    RUNSPACEPOOL_HOST_CALL = 0x00021100
    RUNSPACEPOOL_HOST_RESPONSE = 0x00021101
    PIPELINE_INPUT = 0x00041002
    END_OF_PIPELINE_INPUT = 0x00041003
    PIPELINE_OUTPUT = 0x00041004
    ERROR_RECORD = 0x00041005
    PIPELINE_STATE = 0x00041006
    DEBUG_RECORD = 0x00041007
    VERBOSE_RECORD = 0x00041008
    WARNING_RECORD = 0x00041009
    PROGRESS_RECORD = 0x00041010
    INFORMATION_RECORD = 0x00041011
    PIPELINE_HOST_CALL = 0x00041100
    PIPELINE_HOST_RESPONSE = 0x00041101


class Message(object):

    def __init__(self, destination, rpid, pid, data, serializer):
        """
        [MS-PSRP] 2.2.1 PowerShell Remoting Protocol Message
        https://msdn.microsoft.com/en-us/library/dd303832.aspx

        Used to contain a PSRP message in the structure required by PSRP.

        :param destination: The Destination of the message
        :param rpid: The uuid representation of the RunspacePool
        :param pid: The uuid representation of the PowerShell pipeline
        :param data: The PSRP Message object
        :param serializer: The serializer object used to serialize the message
            when packing
        """
        self.destination = destination
        self.message_type = data.MESSAGE_TYPE

        empty_uuid = uuid.UUID(bytes=b"\x00" * 16)
        self.rpid = uuid.UUID(rpid) if rpid is not None else empty_uuid
        self.pid = uuid.UUID(pid) if pid is not None else empty_uuid
        self.data = data
        self._serializer = serializer

    def pack(self):
        if self.message_type == MessageType.PUBLIC_KEY_REQUEST:
            message_data = ET.Element("S")
        elif self.message_type == MessageType.END_OF_PIPELINE_INPUT:
            message_data = b""
        elif self.message_type == MessageType.PIPELINE_INPUT:
            message_data = self._serializer.serialize(self.data.data)
        elif self.message_type == MessageType.CONNECT_RUNSPACEPOOL and \
                (self.data.min_runspaces is None and
                 self.data.max_runspaces is None):
            message_data = ET.Element("S")
        else:
            message_data = self._serializer.serialize(self.data)

        if not isinstance(message_data, binary_type):
            message_data = \
                ET.tostring(message_data, encoding='utf-8', method='xml')
        log.debug("Packing PSRP message: %s" % to_string(message_data))

        data = struct.pack("<I", self.destination)
        data += struct.pack("<I", self.message_type)

        # .NET stores uuids/guids in bytes in the little endian form
        data += self.rpid.bytes_le
        data += self.pid.bytes_le
        data += message_data

        return data

    @staticmethod
    def unpack(data, serializer):
        destination = struct.unpack("<I", data[0:4])[0]
        message_type = struct.unpack("<I", data[4:8])[0]
        rpid = str(uuid.UUID(bytes_le=data[8:24]))
        pid = str(uuid.UUID(bytes_le=data[24:40]))

        if data[40:43] == b"\xEF\xBB\xBF":
            # 40-43 is the UTF-8 BOM which we don't care about
            message_data = to_string(data[43:])
        else:
            message_data = to_string(data[40:])

        log.debug("Unpacking PSRP message of type %d: %s"
                  % (message_type, message_data))

        message_obj = {
            MessageType.SESSION_CAPABILITY: SessionCapability,
            MessageType.INIT_RUNSPACEPOOL: InitRunspacePool,
            MessageType.PUBLIC_KEY: PublicKey,
            MessageType.ENCRYPTED_SESSION_KEY: EncryptedSessionKey,
            MessageType.PUBLIC_KEY_REQUEST: PublicKeyRequest,
            MessageType.SET_MAX_RUNSPACES: SetMaxRunspaces,
            MessageType.SET_MIN_RUNSPACES: SetMinRunspaces,
            MessageType.RUNSPACE_AVAILABILITY: RunspaceAvailability,
            MessageType.RUNSPACEPOOL_STATE: RunspacePoolStateMessage,
            MessageType.CREATE_PIPELINE: CreatePipeline,
            MessageType.GET_AVAILABLE_RUNSPACES: GetAvailableRunspaces,
            MessageType.USER_EVENT: UserEvent,
            MessageType.APPLICATION_PRIVATE_DATA: ApplicationPrivateData,
            MessageType.GET_COMMAND_METADATA: GetCommandMetadata,
            MessageType.RUNSPACEPOOL_HOST_CALL: RunspacePoolHostCall,
            MessageType.RUNSPACEPOOL_HOST_RESPONSE: RunspacePoolHostResponse,
            MessageType.PIPELINE_INPUT: PipelineInput,
            MessageType.END_OF_PIPELINE_INPUT: EndOfPipelineInput,
            MessageType.PIPELINE_OUTPUT: PipelineOutput,
            MessageType.ERROR_RECORD: ErrorRecordMessage,
            MessageType.PIPELINE_STATE: PipelineState,
            MessageType.DEBUG_RECORD: DebugRecord,
            MessageType.VERBOSE_RECORD: VerboseRecord,
            MessageType.WARNING_RECORD: WarningRecord,
            MessageType.PROGRESS_RECORD: ProgressRecord,
            MessageType.INFORMATION_RECORD: InformationRecord,
            MessageType.PIPELINE_HOST_CALL: PipelineHostCall,
            MessageType.PIPELINE_HOST_RESPONSE: PipelineHostResponse,
            MessageType.CONNECT_RUNSPACEPOOL: ConnectRunspacePool,
            MessageType.RUNSPACEPOOL_INIT_DATA: RunspacePoolInitData,
            MessageType.RESET_RUNSPACE_STATE: ResetRunspaceState
        }[message_type]

        # PIPELINE_OUTPUT is a weird one, it contains the actual output objects
        # not encapsulated so we set it to a dynamic object and the serializer
        # will work out what is best
        if message_type == MessageType.PIPELINE_OUTPUT:
            # try to deserialize using our known objects, if that fails then
            # we want to get a generic object at least but raise a warning
            try:
                message_data = serializer.deserialize(message_data)
            except SerializationError as err:
                warnings.warn("Failed to deserialize msg, trying to "
                              "deserialize as generic complex object: %s"
                              % str(err))
                meta = ObjectMeta("ObjDynamic", object=GenericComplexObject)
                message_data = serializer.deserialize(message_data, meta)
            message = PipelineOutput()
            message.data = message_data
        elif message_type == MessageType.PIPELINE_INPUT:
            message_data = serializer.deserialize(message_data)
            message = PipelineInput()
            message.data = message_data
        elif message_type == MessageType.PUBLIC_KEY_REQUEST:
            message = PublicKeyRequest()
        else:
            message_meta = ObjectMeta("Obj", object=message_obj)
            message = serializer.deserialize(message_data, message_meta)

        return Message(destination, rpid, pid, message, serializer)


class SessionCapability(ComplexObject):
    MESSAGE_TYPE = MessageType.SESSION_CAPABILITY

    def __init__(self, protocol_version=None, ps_version=None,
                 serialization_version=None, time_zone=None):
        """
        [MS-PSRP] 2.2.2.1 SESSION_CAPABILITY Message
        https://msdn.microsoft.com/en-us/library/dd340636.aspx

        :param protocol_version: The PSRP version
        :param ps_version: The PowerShell version
        :param serialization_version: The serialization version
        :param time_zone: Time Zone information of the host, should be a byte
            string
        """
        super(SessionCapability, self).__init__()
        self._extended_properties = (
            ('protocol_version', ObjectMeta("Version",
                                            name="protocolversion")),
            ('ps_version', ObjectMeta("Version", name="PSVersion")),
            ('serialization_version', ObjectMeta("Version",
                                                 name="SerializationVersion")),
            ('time_zone', ObjectMeta("BA", name="TimeZone", optional=True)),
        )
        self.protocol_version = protocol_version
        self.ps_version = ps_version
        self.serialization_version = serialization_version
        self.time_zone = time_zone


class InitRunspacePool(ComplexObject):
    MESSAGE_TYPE = MessageType.INIT_RUNSPACEPOOL

    def __init__(self, min_runspaces=None, max_runspaces=None,
                 thread_options=None, apartment_state=None, host_info=None,
                 application_arguments=None):
        """
        [MS-PSRP] 2.2.2.2 INIT_RUNSPACEPOOL Message
        https://msdn.microsoft.com/en-us/library/dd359645.aspx

        :param min_runspaces: The minimum number of runspaces in the pool
        :param max_runspaces: The maximum number of runspaces in the pool
        :param thread_options: Thread options provided by the higher layer
        :param apartment_state: Apartment state provided by the higher layer
        :param host_info: The client's HostInfo details
        :param application_arguments: Application arguments provided by a
            higher layer, stored in the $PSSenderInfo variable in the pool
        """
        super(InitRunspacePool, self).__init__()
        self._extended_properties = (
            ('min_runspaces', ObjectMeta("I32", name="MinRunspaces")),
            ('max_runspaces', ObjectMeta("I32", name="MaxRunspaces")),
            ('thread_options', ObjectMeta("Obj", name="PSThreadOptions",
                                          object=PSThreadOptions)),
            ('apartment_state', ObjectMeta("Obj", name="ApartmentState",
                                           object=ApartmentState)),
            ('host_info', ObjectMeta("Obj", name="HostInfo",
                                     object=HostInfo)),
            ('application_arguments', DictionaryMeta(
                name="ApplicationArguments",
                dict_types=[
                    "System.Management.Automation.PSPrimitiveDictionary",
                    "System.Collections.Hashtable",
                    "System.Object"
                ]
            ))
        )
        self.min_runspaces = min_runspaces
        self.max_runspaces = max_runspaces
        self.thread_options = thread_options
        self.apartment_state = apartment_state
        self.host_info = host_info
        self.application_arguments = application_arguments


class PublicKey(ComplexObject):
    MESSAGE_TYPE = MessageType.PUBLIC_KEY

    def __init__(self, public_key=None):
        """
        [MS-PSRP] 2.2.2.3 PUBLIC_KEY Message
        https://msdn.microsoft.com/en-us/library/dd644859.aspx

        :param public_key: The Base64 encoding of the public key in the PKCS1
            format.
        """
        super(PublicKey, self).__init__()
        self._extended_properties = (
            ('public_key', ObjectMeta("S", name="PublicKey")),
        )
        self.public_key = public_key


class EncryptedSessionKey(ComplexObject):
    MESSAGE_TYPE = MessageType.ENCRYPTED_SESSION_KEY

    def __init__(self, session_key=None):
        """
        [MS-PSRP] 2.2.2.4 ENCRYPTED_SESSION_KEY Message
        https://msdn.microsoft.com/en-us/library/dd644930.aspx

        :param session_key: The 256-bit key for AES encryption that has been
            encrypted using the public key from the PUBLIC_KEY message using
            the RSAES-PKCS-v1_5 encryption scheme and then Base64 formatted.
        """
        super(EncryptedSessionKey, self).__init__()
        self._extended_properties = (
            ('session_key', ObjectMeta("S", name="EncryptedSessionKey")),
        )
        self.session_key = session_key


class PublicKeyRequest(object):
    MESSAGE_TYPE = MessageType.PUBLIC_KEY_REQUEST

    def __init__(self):
        """
        [MS-PSRP] 2.2.2.5 PUBLIC_KEY_REQUEST Message
        https://msdn.microsoft.com/en-us/library/dd644906.aspx
        """
        super(PublicKeyRequest, self).__init__()


class SetMaxRunspaces(ComplexObject):
    MESSAGE_TYPE = MessageType.SET_MAX_RUNSPACES

    def __init__(self, max_runspaces=None, ci=None):
        """
        [MS-PSRP] 2.2.2.6 SET_MAX_RUNSPACES Message
        https://msdn.microsoft.com/en-us/library/dd304870.aspx

        :param max_runspaces: The maximum number of runspaces
        :param ci: The ci identifier for the CI table
        """
        super(SetMaxRunspaces, self).__init__()
        self._extended_properties = (
            ('max_runspaces', ObjectMeta("I32", name="MaxRunspaces")),
            ('ci', ObjectMeta("I64", name="CI")),
        )
        self.max_runspaces = max_runspaces
        self.ci = ci


class SetMinRunspaces(ComplexObject):
    MESSAGE_TYPE = MessageType.SET_MIN_RUNSPACES

    def __init__(self, min_runspaces=None, ci=None):
        """
        [MS-PSRP] 2.2.2.7 SET_MIN_RUNSPACES Message
        https://msdn.microsoft.com/en-us/library/dd340570.aspx

        :param max_runspaces: The minimum number of runspaces
        :param ci: The ci identifier for the CI table
        """
        super(SetMinRunspaces, self).__init__()
        self._extended_properties = (
            ('min_runspaces', ObjectMeta("I32", name="MinRunspaces")),
            ('ci', ObjectMeta("I64", name="CI")),
        )
        self.min_runspaces = min_runspaces
        self.ci = ci


class RunspaceAvailability(ComplexObject):
    MESSAGE_TYPE = MessageType.RUNSPACE_AVAILABILITY

    def __init__(self, response=None, ci=None):
        """
        [MS-PSRP] 2.2.2.8 RUNSPACE_AVAILABILITY Message
        https://msdn.microsoft.com/en-us/library/dd359229.aspx

        :param response: The response from the server
        :param ci: The ci identifier for the CI table
        """
        super(RunspaceAvailability, self).__init__()
        self._extended_properties = (
            ('response', ObjectMeta(name="SetMinMaxRunspacesResponse")),
            ('ci', ObjectMeta("I64", name="ci")),
        )
        self.response = response
        self.ci = ci


class RunspacePoolStateMessage(ComplexObject):
    MESSAGE_TYPE = MessageType.RUNSPACEPOOL_STATE

    def __init__(self, state=None, error_record=None):
        """
        [MS-PSRP] 2.2.2.9 RUNSPACEPOOL_STATE Message
        https://msdn.microsoft.com/en-us/library/dd303020.aspx

        :param state: The state of the runspace pool
        :param error_record:
        """
        super(RunspacePoolStateMessage, self).__init__()
        self._extended_properties = (
            ('state', ObjectMeta("I32", name="RunspaceState")),
            ('error_record', ObjectMeta("Obj", optional=True,
                                        object=ErrorRecord)),
        )
        self.state = state
        self.error_record = error_record


class CreatePipeline(ComplexObject):
    MESSAGE_TYPE = MessageType.CREATE_PIPELINE

    def __init__(self, no_input=None, apartment_state=None,
                 remote_stream_options=None, add_to_history=None,
                 host_info=None, pipeline=None, is_nested=None):
        """
        [MS-PSRP] 2.2.2.10 CREATE_PIPELINE Message
        https://msdn.microsoft.com/en-us/library/dd340567.aspx

        :param no_input: Whether the pipeline will take input
        :param apartment_state: The ApartmentState of the pipeline
        :param remote_stream_options: The RemoteStreamOptions of the pipeline
        :param add_to_history: Whether to add the pipeline being execute to
            the history field of the runspace
        :param host_info: The HostInformation of the pipeline
        :param pipeline: The PowerShell object to create
        :param is_nested: Whether the pipeline is run in nested or
            steppable mode
        """
        super(CreatePipeline, self).__init__()
        self._extended_properties = (
            ('no_input', ObjectMeta("B", name="NoInput")),
            ('apartment_state', ObjectMeta("Obj", name="ApartmentState",
                                           object=ApartmentState)),
            ('remote_stream_options', ObjectMeta("Obj",
                                                 name="RemoteStreamOptions",
                                                 object=RemoteStreamOptions)),
            ('add_to_history', ObjectMeta("B", name="AddToHistory")),
            ('host_info', ObjectMeta("Obj", name="HostInfo",
                                     object=HostInfo)),
            ('pipeline', ObjectMeta("Obj", name="PowerShell",
                                    object=Pipeline)),
            ('is_nested', ObjectMeta("B", name="IsNested")),
        )

        self.no_input = no_input
        self.apartment_state = apartment_state
        self.remote_stream_options = remote_stream_options
        self.add_to_history = add_to_history
        self.host_info = host_info
        self.pipeline = pipeline
        self.is_nested = is_nested


class GetAvailableRunspaces(ComplexObject):
    MESSAGE_TYPE = MessageType.GET_AVAILABLE_RUNSPACES

    def __init__(self, ci=None):
        """
        [MS-PSRP] 2.2.2.11 GET_AVAILABLE_RUNSPACES Message
        https://msdn.microsoft.com/en-us/library/dd357512.aspx

        :param ci: The ci identifier for the CI table
        """
        super(GetAvailableRunspaces, self).__init__()
        self._extended_properties = (
            ('ci', ObjectMeta("I64", name="ci")),
        )
        self.ci = ci


class UserEvent(ComplexObject):
    MESSAGE_TYPE = MessageType.USER_EVENT

    def __init__(self, event_id=None, source_id=None, time=None, sender=None,
                 args=None, data=None, computer=None, runspace_id=None):
        """
        [MS-PSRP] 2.2.2.12 USER_EVENT Message
        https://msdn.microsoft.com/en-us/library/dd359395.aspx

        :param event_id:
        :param source_id:
        :param time:
        :param sender:
        :param args:
        :param data:
        :param computer:
        :param runspace_id:
        """
        super(UserEvent, self).__init__()
        self._extended_properties = (
            ('event_id', ObjectMeta(
                "I32", name="PSEventArgs.EventIdentifier")),
            ('source_id', ObjectMeta(
                "S", name="PSEventArgs.SourceIdentifier")),
            ('time', ObjectMeta("DT", name="PSEventArgs.TimeGenerated")),
            ('sender', ObjectMeta(name="PSEventArgs.Sender")),
            ('args', ObjectMeta(name="PSEventArgs.SourceArgs")),
            ('data', ObjectMeta(name="PSEventArgs.MessageData")),
            ('computer', ObjectMeta("S", name="PSEventArgs.ComputerName")),
            ('runspace_id', ObjectMeta("G", name="PSEventArgs.RunspaceId")),
        )
        self.event_id = event_id
        self.source_id = source_id
        self.time = time
        self.sender = sender
        self.args = args
        self.data = data
        self.computer = computer
        self.runspace_id = runspace_id


class ApplicationPrivateData(ComplexObject):
    MESSAGE_TYPE = MessageType.APPLICATION_PRIVATE_DATA

    def __init__(self, data=None):
        """
        [MS-PSRP] 2.2.2.13 APPLICATION_PRIVATE_DATA Message
        https://msdn.microsoft.com/en-us/library/dd644934.aspx

        :param data: A dict that contains data to sent to the PowerShell layer
        """
        super(ApplicationPrivateData, self).__init__()
        self._extended_properties = (
            ('data', DictionaryMeta(name="ApplicationPrivateData")),
        )
        self.data = data


class GetCommandMetadata(ComplexObject):
    MESSAGE_TYPE = MessageType.GET_COMMAND_METADATA

    def __init__(self, names=None, command_type=None, namespace=None,
                 argument_list=None):
        """
        [MS-PSRP] 2.2.2.14 GET_COMMAND_METADATA Message
        https://msdn.microsoft.com/en-us/library/ee175985.aspx

        :param names:
        :param command_type:
        :param namespace:
        :param argument_list:
        """
        super(GetCommandMetadata, self).__init__()
        self._extended_properties = (
            ('names', ListMeta(
                name="Name", list_value_meta=ObjectMeta("S"),
                list_types=[
                    "System.String[]",
                    "System.Array",
                    "System.Object"
                ])
             ),
            ('command_type', ObjectMeta(name="CommandType",
                                        object=CommandType)),
            ('namespace', ObjectMeta(name="Namespace")),
            ('argument_list', ListMeta(name="ArgumentList"))
        )
        self.names = names
        self.command_type = command_type
        self.namespace = namespace
        self.argument_list = argument_list


class RunspacePoolHostCall(ComplexObject):
    MESSAGE_TYPE = MessageType.RUNSPACEPOOL_HOST_CALL

    def __init__(self, ci=None, mi=None, mp=None):
        """
        [MS-PSRP] 2.2.2.15 RUNSPACE_HOST_CALL Message
        https://msdn.microsoft.com/en-us/library/dd340830.aspx

        :param ci:
        :param mi:
        :param mp:
        """
        super(RunspacePoolHostCall, self).__init__()
        self._extended_properties = (
            ('ci', ObjectMeta("I64", name="ci")),
            ('mi', ObjectMeta("Obj", name="mi", object=HostMethodIdentifier)),
            ('mp', ListMeta(name="mp"))
        )
        self.ci = ci
        self.mi = mi
        self.mp = mp


class RunspacePoolHostResponse(ComplexObject):
    MESSAGE_TYPE = MessageType.RUNSPACEPOOL_HOST_RESPONSE

    def __init__(self, ci=None, mi=None, mr=None, me=None):
        """
        [MS-PSRP] 2.2.2.16 RUNSPACEPOOL_HOST_RESPONSE Message
        https://msdn.microsoft.com/en-us/library/dd358453.aspx

        :param ci:
        :param mi:
        :param mr:
        :param me:
        """
        super(RunspacePoolHostResponse, self).__init__()
        self._extended_properties = (
            ('ci', ObjectMeta("I64", name="ci")),
            ('mi', ObjectMeta("Obj", name="mi", object=HostMethodIdentifier)),
            ('mr', ObjectMeta(name="mr")),
            ('me', ObjectMeta("Obj", name="me", object=ErrorRecord,
                              optional=True)),
        )
        self.ci = ci
        self.mi = mi
        self.mr = mr
        self.me = me


class PipelineInput(ComplexObject):
    MESSAGE_TYPE = MessageType.PIPELINE_INPUT

    def __init__(self, data=None):
        """
        [MS-PSRP] 2.2.2.17 PIPELINE_INPUT Message
        https://msdn.microsoft.com/en-us/library/dd340525.aspx

        :param data: The data to serialize and send as the input
        """
        super(PipelineInput, self).__init__()
        self.data = data


class EndOfPipelineInput(ComplexObject):
    MESSAGE_TYPE = MessageType.END_OF_PIPELINE_INPUT

    def __init__(self):
        """
        [MS-PSRP] 2.2.2.18 END_OF_PIPELINE_INPUT Message
        https://msdn.microsoft.com/en-us/library/dd342785.aspx
        """
        super(EndOfPipelineInput, self).__init__()


class PipelineOutput(object):
    MESSAGE_TYPE = MessageType.PIPELINE_OUTPUT

    def __init__(self, data=None):
        """
        [MS-PSRP] 2.2.2.19 PIPELINE_OUTPUT Message
        https://msdn.microsoft.com/en-us/library/dd357371.aspx
        """
        super(PipelineOutput, self).__init__()
        self.data = data


class ErrorRecordMessage(ErrorRecord):
    MESSAGE_TYPE = MessageType.ERROR_RECORD

    def __init__(self, **kwargs):
        """
        [MS-PSRP] 2.2.2.20 ERROR_RECORD Message
        https://msdn.microsoft.com/en-us/library/dd342423.aspx

        :param kwargs:
        """
        super(ErrorRecordMessage, self).__init__(**kwargs)


class PipelineState(ComplexObject):
    MESSAGE_TYPE = MessageType.PIPELINE_STATE

    def __init__(self, state=None, error_record=None):
        """
        [MS-PSRP] 2.2.2.21 PIPELINE_STATE Message
        https://msdn.microsoft.com/en-us/library/dd304923.aspx

        :param state: The state of the pipeline
        :param error_record:
        """
        super(PipelineState, self).__init__()
        self._extended_properties = (
            ('state', ObjectMeta("I32", name="PipelineState")),
            ('error_record', ObjectMeta("Obj", name="ExceptionAsErrorRecord",
                                        optional=True)),
        )
        self.state = state
        self.error_record = error_record


class DebugRecord(InformationalRecord):
    MESSAGE_TYPE = MessageType.DEBUG_RECORD

    def __init__(self, **kwargs):
        """
        [MS-PSRP] 2.2.2.22 DEBUG_RECORD Message
        https://msdn.microsoft.com/en-us/library/dd340758.aspx
        """
        super(DebugRecord, self).__init__(**kwargs)
        self._types.insert(0, "System.Management.Automation.DebugRecord")


class VerboseRecord(InformationalRecord):
    MESSAGE_TYPE = MessageType.VERBOSE_RECORD

    def __init__(self, **kwargs):
        """
        [MS-PSRP] 2.2.2.23 VERBOSE_RECORD Message
        https://msdn.microsoft.com/en-us/library/dd342930.aspx
        """
        super(VerboseRecord, self).__init__(**kwargs)
        self._types.insert(0, "System.Management.Automation.VerboseRecord")


class WarningRecord(InformationalRecord):
    MESSAGE_TYPE = MessageType.WARNING_RECORD

    def __init__(self, **kwargs):
        """
        [MS-PSRP] 2.2.2.24 WARNING_RECORD Message
        https://msdn.microsoft.com/en-us/library/dd303590.aspx
        """
        super(WarningRecord, self).__init__(**kwargs)
        self._types.insert(0, "System.Management.Automation.WarningRecord")


class ProgressRecord(ComplexObject):
    MESSAGE_TYPE = MessageType.PROGRESS_RECORD

    def __init__(self, activity=None, activity_id=None, description=None,
                 current_operation=None, parent_activity_id=None,
                 percent_complete=None, progress_type=None,
                 seconds_remaining=None):
        """
        [MS-PSRP] 2.2.2.25 PROGRESS_RECORD Message
        https://msdn.microsoft.com/en-us/library/dd340751.aspx

        :param kwargs:
        """
        super(ProgressRecord, self).__init__()
        self._extended_properties = (
            ('activity', ObjectMeta("S", name="Activity")),
            ('activity_id', ObjectMeta("I32", name="ActivityId")),
            ('description', ObjectMeta("S", name="StatusDescription")),
            ('current_operation', ObjectMeta("S", name="CurrentOperation")),
            ('parent_activity_id', ObjectMeta("I32", name="ParentActivityId")),
            ('percent_complete', ObjectMeta("I32", name="PercentComplete")),
            ('progress_type', ObjectMeta("Obj", name="Type",
                                         object=ProgressRecordType)),
            ('seconds_remaining', ObjectMeta("I32", name="SecondsRemaining")),
        )
        self.activity = activity
        self.activity_id = activity_id
        self.description = description
        self.current_operation = current_operation
        self.parent_activity_id = parent_activity_id
        self.percent_complete = percent_complete
        self.progress_type = progress_type
        self.seconds_remaining = seconds_remaining


class InformationRecord(ComplexObject):
    MESSAGE_TYPE = MessageType.INFORMATION_RECORD

    def __init__(self, message_data=None, source=None, time_generated=None,
                 tags=None, user=None, computer=None, pid=None,
                 native_thread_id=None, managed_thread_id=None,
                 write_information_stream=None):
        """
        [MS-PSRP] 2.2.2.26 INFORMATION_RECORD Message
        https://msdn.microsoft.com/en-us/library/mt224023.aspx

        Only in protocol_version 2.3 and above

        :param kwargs:
        """
        super(InformationRecord, self).__init__()
        self._types = [
            "System.Management.Automation.InformationRecord",
            "System.Object"
        ]
        self._extended_properties = (
            ('message_data', ObjectMeta(name="MessageData")),
            ('source', ObjectMeta("S", name="Source")),
            ('time_generated', ObjectMeta("DT", name="TimeGenerated")),
            ('tags', ListMeta(name="Tags", list_value_meta=ObjectMeta("S"))),
            ('user', ObjectMeta("S", name="User")),
            ('computer', ObjectMeta("S", name="Computer")),
            ('pid', ObjectMeta("U32", name="ProcessId")),
            ('native_thread_id', ObjectMeta("U32", name="NativeThreadId")),
            ('managed_thread_id', ObjectMeta("U32", name="ManagedThreadId")),
            ('write_information_stream', ObjectMeta(
                "B", name="WriteInformationStream", optional=True
            )),
        )
        self.message_data = message_data
        self.source = source
        self.time_generated = time_generated
        self.tags = tags
        self.user = user
        self.computer = computer
        self.pid = pid
        self.native_thread_id = native_thread_id
        self.managed_thread_id = managed_thread_id
        self.write_information_stream = write_information_stream


class PipelineHostCall(RunspacePoolHostCall):
    MESSAGE_TYPE = MessageType.PIPELINE_HOST_CALL
    """
    [MS-PSRP] 2.2.2.27 PIPELINE_HOST_CALL Message
    https://msdn.microsoft.com/en-us/library/dd356915.aspx
    """


class PipelineHostResponse(RunspacePoolHostResponse):
    MESSAGE_TYPE = MessageType.PIPELINE_HOST_RESPONSE
    """
    [MS-PSRP] 2.2.2.28 PIPELINE_HOST_RESPONSE Message
    https://msdn.microsoft.com/en-us/library/dd306168.aspx
    """


class ConnectRunspacePool(ComplexObject):
    MESSAGE_TYPE = MessageType.CONNECT_RUNSPACEPOOL

    def __init__(self, min_runspaces=None, max_runspaces=None):
        """
        [MS-PSRP] 2.2.2.29 CONNECT_RUNSPACEPOOL Message
        https://msdn.microsoft.com/en-us/library/hh537460.aspx

        :param min_runspaces:
        :param max_runspaces:
        """
        super(ConnectRunspacePool, self).__init__()
        self._extended_properties = (
            ('min_runspaces', ObjectMeta("I32", name="MinRunspaces",
                                         optional=True)),
            ('max_runspaces', ObjectMeta("I32", name="MaxRunspaces",
                                         optional=True)),
        )
        self.min_runspaces = min_runspaces
        self.max_runspaces = max_runspaces


class RunspacePoolInitData(ComplexObject):
    MESSAGE_TYPE = MessageType.RUNSPACEPOOL_INIT_DATA

    def __init__(self, min_runspaces=None, max_runspaces=None):
        """
        [MS-PSRP] 2.2.2.30 RUNSPACEPOOL_INIT_DATA Message
        https://msdn.microsoft.com/en-us/library/hh537788.aspx

        :param min_runspaces:
        :param max_runspaces:
        """
        super(RunspacePoolInitData, self).__init__()
        self._extended_properties = (
            ('min_runspaces', ObjectMeta("I32", name="MinRunspaces")),
            ('max_runspaces', ObjectMeta("I32", name="MaxRunspaces")),
        )
        self.min_runspaces = min_runspaces
        self.max_runspaces = max_runspaces


class ResetRunspaceState(ComplexObject):
    MESSAGE_TYPE = MessageType.RESET_RUNSPACE_STATE

    def __init__(self, ci=None):
        """
        [MS-PSRP] 2.2.2.31 RESET_RUNSPACE_STATE Message
        https://msdn.microsoft.com/en-us/library/mt224027.aspx

        :param ci: The call identifier
        """
        super(ResetRunspaceState, self).__init__()
        self._extended_properties = (
            ('ci', ObjectMeta("I64", name="ci")),
        )
        self.ci = ci



import logging
import re
import struct

from pypsrp._utils import to_bytes
from pypsrp.exceptions import WinRMError

log = logging.getLogger(__name__)


class WinRMEncryption(object):

    SIXTEEN_KB = 16384
    MIME_BOUNDARY = "--Encrypted Boundary"
    CREDSSP = "application/HTTP-CredSSP-session-encrypted"
    KERBEROS = "application/HTTP-Kerberos-session-encrypted"
    SPNEGO = "application/HTTP-SPNEGO-session-encrypted"

    def __init__(self, context, protocol):
        log.debug("Initialising WinRMEncryption helper for protocol %s" % protocol)
        self.context = context
        self.protocol = protocol

        if protocol == self.CREDSSP:
            self._wrap = self._wrap_credssp
            self._unwrap = self._unwrap_credssp
        else:
            self._wrap = self._wrap_spnego
            self._unwrap = self._unwrap_spnego

    def wrap_message(self, message):
        log.debug("Wrapping message")
        if self.protocol == self.CREDSSP and len(message) > self.SIXTEEN_KB:
            content_type = "multipart/x-multi-encrypted"
            encrypted_msg = b""
            chunks = [message[i:i + self.SIXTEEN_KB] for i in
                      range(0, len(message), self.SIXTEEN_KB)]
            for chunk in chunks:
                encrypted_chunk = self._wrap_message(chunk)
                encrypted_msg += encrypted_chunk
        else:
            content_type = "multipart/encrypted"
            encrypted_msg = self._wrap_message(message)

        encrypted_msg += to_bytes("%s--\r\n" % self.MIME_BOUNDARY)

        log.debug("Created wrapped message of content type %s" % content_type)
        return content_type, encrypted_msg

    def unwrap_message(self, message, boundary):
        log.debug("Unwrapped message")

        # Talking to Exchange endpoints gives a non-compliant boundary that has a space between the -- {boundary}, not
        # ideal but we just need to handle it.
        parts = re.compile(to_bytes(r"--\s*%s\r\n" % re.escape(boundary))).split(message)
        parts = list(filter(None, parts))

        message = b""
        for i in range(0, len(parts), 2):
            header = parts[i].strip()
            payload = parts[i + 1]

            expected_length = int(header.split(b"Length=")[1])

            # remove the end MIME block if it exists
            payload = re.sub(to_bytes(r'--\s*%s--\r\n$') % to_bytes(boundary), b'', payload)

            wrapped_data = payload.replace(b"\tContent-Type: application/octet-stream\r\n", b"")
            unwrapped_data = self._unwrap(wrapped_data)
            actual_length = len(unwrapped_data)

            log.debug("Actual unwrapped length: %d, expected unwrapped length:"
                      " %d" % (actual_length, expected_length))
            if actual_length != expected_length:
                raise WinRMError("The encrypted length from the server does "
                                 "not match the expected length, decryption "
                                 "failed, actual: %d != expected: %d"
                                 % (actual_length, expected_length))
            message += unwrapped_data

        return message

    def _wrap_message(self, message):
        wrapped_data, padding_length = self._wrap(message)
        msg_length = str(len(message) + padding_length)

        payload = "\r\n".join([
            self.MIME_BOUNDARY,
            "\tContent-Type: %s" % self.protocol,
            "\tOriginalContent: type=application/soap+xml;charset=UTF-8;Length=%s" % msg_length,
            self.MIME_BOUNDARY,
            "\tContent-Type: application/octet-stream",
            ""
        ])
        payload = to_bytes(payload) + wrapped_data

        return payload

    def _wrap_spnego(self, data):
        header, wrapped_data, padding_length = self.context.wrap_winrm(data)

        return struct.pack("<i", len(header)) + header + wrapped_data, padding_length

    def _wrap_credssp(self, data):
        wrapped_data = self.context.wrap(data)
        cipher_negotiated = self.context.tls_connection.get_cipher_name()
        trailer_length = self._credssp_trailer(len(data), cipher_negotiated)

        return struct.pack("<i", trailer_length) + wrapped_data, 0

    def _unwrap_spnego(self, data):
        header_length = struct.unpack("<i", data[:4])[0]
        header = data[4:4 + header_length]
        wrapped_data = data[4 + header_length:]

        data = self.context.unwrap_winrm(header, wrapped_data)

        return data

    def _unwrap_credssp(self, data):
        wrapped_data = data[4:]
        data = self.context.unwrap(wrapped_data)

        return data

    def _credssp_trailer(self, msg_len, cipher_suite):
        # On Windows this is derived from SecPkgContext_StreamSizes, this is
        # not available on other platforms so we need to calculate it manually
        log.debug("Attempting to get CredSSP trailer length for msg of "
                  "length %d with cipher %s" % (msg_len, cipher_suite))

        if re.match(r'^.*-GCM-[\w\d]*$', cipher_suite):
            # GCM has a fixed length of 16 bytes
            trailer_length = 16
        else:
            # For other cipher suites, trailer size == len(hmac) + len(padding)
            # the padding it the length required by the chosen block cipher
            hash_algorithm = cipher_suite.split('-')[-1]

            # while there are other algorithms, SChannel doesn't support them
            # as of yet so we just keep to this list
            hash_length = {
                'MD5': 16,
                'SHA': 20,
                'SHA256': 32,
                'SHA384': 48
            }.get(hash_algorithm, 0)

            pre_pad_length = msg_len + hash_length
            if "RC4" in cipher_suite:
                # RC4 is a stream cipher so no padding would be added
                padding_length = 0
            elif "DES" in cipher_suite or "3DES" in cipher_suite:
                # 3DES is a 64 bit block cipher
                padding_length = 8 - (pre_pad_length % 8)
            else:
                # AES is a 128 bit block cipher
                padding_length = 16 - (pre_pad_length % 16)

            trailer_length = (pre_pad_length + padding_length) - msg_len

        return trailer_length



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright: (c) 2020 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import os

from setuptools import setup


def abs_path(rel_path):
    return os.path.join(os.path.dirname(__file__), rel_path)


with open(abs_path('README.md'), mode='rb') as fd:
    long_description = fd.read().decode('utf-8')


setup(
    name='pypsrp',
    version='0.5.0',
    packages=['pypsrp', 'pypsrp.pwsh_scripts'],
    include_package_data=True,
    install_requires=[
        'cryptography',
        'pyspnego',
        'requests>=2.9.1',
        'six',
    ],
    extras_require={
        ':python_version<="2.7"': [
            'ipaddress',
        ],
        'credssp': [
            'requests-credssp>=1.0.0'
        ],
        'kerberos:sys_platform=="win32"': [],
        'kerberos:sys_platform!="win32"': [
            'gssapi>=1.5.0'
        ]
    },
    author='Jordan Borean',
    author_email='jborean93@gmail.com',
    url='https://github.com/jborean93/pypsrp',
    description='PowerShell Remoting Protocol and WinRM for Python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='winrm psrp winrs windows powershell',
    license='MIT',
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



# Copyright: (c) 2018, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import division

import base64
import hashlib
import logging
import os
import shutil
import sys
import tempfile
import warnings

from pypsrp.exceptions import WinRMError
from pypsrp.powershell import DEFAULT_CONFIGURATION_NAME, PowerShell, RunspacePool
from pypsrp.serializer import Serializer
from pypsrp.shell import Process, SignalCode, WinRS
from pypsrp.wsman import WSMan
from pypsrp._utils import get_pwsh_script, to_bytes, to_unicode

if sys.version_info[0] == 2 and sys.version_info[1] < 7:  # pragma: no cover
    # ElementTree in Python 2.6 does not support namespaces so we need to use
    # lxml instead for this version
    from lxml import etree as ET
else:  # pragma: no cover
    import xml.etree.ElementTree as ET

log = logging.getLogger(__name__)


class Client(object):

    def __init__(self, server, **kwargs):
        """
        Creates a client object used to do the following
            spawn new cmd command/process
            spawn new PowerShell Runspace Pool/Pipeline
            copy a file from localhost to the remote Windows host
            fetch a file from the remote Windows host to localhost

        This is just an easy to use layer on top of the objects WinRS and
        RunspacePool/PowerShell. It trades flexibility in favour of simplicity.

        If your use case needs some of that flexibility you can use these
        functions as a reference implementation for your own functions.

        :param server: The server/host to connect to
        :param kwargs: The various WSMan args to control the transport
            mechanism, see pypsrp.wsman.WSMan for these args
        """
        self.wsman = WSMan(server, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def copy(self, src, dest, configuration_name=DEFAULT_CONFIGURATION_NAME,
             expand_variables=False):
        """
        Copies a single file from the current host to the remote Windows host.
        This can be quite slow when it comes to large files due to the
        limitations of WinRM but it is designed to be as fast as it can be.
        During the copy process, the bytes will be stored in a temporary file
        before being copied.

        When copying it will replace the file at dest if one already exists. It
        also will verify the checksum of the copied file is the same as the
        actual file locally before copying the file to the path at dest.

        :param src: The path to the local file
        :param dest: The path to the destination file on the Windows host
        :param configuration_name: The PowerShell configuration endpoint to
            use when copying the file.
        :param expand_variables: Expand variables in path. Disabled by default
            Enable for cmd like expansion (for example %TMP% in path)
        :return: The absolute path of the file on the Windows host
        """

        def read_buffer(b_path, total_size, buffer_size):
            offset = 0
            sha1 = hashlib.sha1()

            with open(b_path, 'rb') as src_file:
                for data in iter((lambda: src_file.read(buffer_size)), b""):
                    log.debug("Reading data of file at offset=%d with size=%d"
                              % (offset, buffer_size))
                    offset += len(data)
                    sha1.update(data)
                    b64_data = base64.b64encode(data)

                    result = [to_unicode(b64_data)]
                    if offset == total_size:
                        result.append(to_unicode(base64.b64encode(to_bytes(sha1.hexdigest()))))

                    yield result

                # the file was empty, return empty buffer
                if offset == 0:
                    yield [u"", to_unicode(base64.b64encode(to_bytes(sha1.hexdigest())))]

        if expand_variables:
            src = os.path.expanduser(os.path.expandvars(src))
        b_src = to_bytes(src)
        src_size = os.path.getsize(b_src)
        log.info("Copying '%s' to '%s' with a total size of %d"
                 % (src, dest, src_size))

        with RunspacePool(self.wsman, configuration_name=configuration_name) as pool:
            # Get the buffer size of each fragment to send, subtract. Adjust to size of the base64 encoded bytes. Also
            # subtract 82 for the fragment, message, and other header info that PSRP adds.
            buffer_size = int((self.wsman.max_payload_size - 82) / 4 * 3)

            log.info("Creating file reader with a buffer size of %d" % buffer_size)
            read_gen = read_buffer(b_src, src_size, buffer_size)

            command = get_pwsh_script('copy.ps1')
            log.debug("Starting to send file data to remote process")
            powershell = PowerShell(pool)
            powershell.add_script(command) \
                .add_argument(dest) \
                .add_argument(expand_variables)
            powershell.invoke(input=read_gen)
            _handle_powershell_error(powershell, "Failed to copy file")

        log.debug("Finished sending file data to remote process")
        for warning in powershell.streams.warning:
            warnings.warn(str(warning))

        output_file = to_unicode(powershell.output[-1]).strip()
        log.info("Completed file transfer of '%s' to '%s'" % (src, output_file))
        return output_file

    def execute_cmd(self, command, encoding='437', environment=None):
        """
        Executes a command in a cmd shell and returns the stdout/stderr/rc of
        that process. This uses the raw WinRS layer and can be used to execute
        a traditional process.

        :param command: The command to execute
        :param encoding: The encoding of the output std buffers, this
            correlates to the codepage of the host and traditionally en-US
            is 437. This probably doesn't need to be modified unless you are
            running a different codepage on your host
        :param environment: A dictionary containing environment keys and
            values to set on the executing process.
        :return: A tuple of
            stdout: A unicode string of the stdout
            stderr: A unicode string of the stderr
            rc: The return code of the process

        Both stdout and stderr are returned from the server as a byte string,
        they are converted to a unicode string based on the encoding variable
        set
        """
        log.info("Executing cmd process '%s'" % command)
        with WinRS(self.wsman, environment=environment) as shell:
            process = Process(shell, command)
            process.invoke()
            process.signal(SignalCode.CTRL_C)

        return to_unicode(process.stdout, encoding), \
            to_unicode(process.stderr, encoding), process.rc

    def execute_ps(self, script, configuration_name=DEFAULT_CONFIGURATION_NAME,
                   environment=None):
        """
        Executes a PowerShell script in a PowerShell runspace pool. This uses
        the PSRP layer and is designed to run a PowerShell script and not a
        raw executable.

        Because this runs in a runspace, traditional concepts like stdout,
        stderr, rc's are no longer relevant. Instead there is a output,
        error/verbose/debug streams, and a boolean that indicates if the
        script execution came across an error. If you want the traditional
        stdout/stderr/rc, use execute_cmd instead.

        :param script: The PowerShell script to run
        :param configuration_name: The PowerShell configuration endpoint to
            use when executing the script.
        :param environment: A dictionary containing environment keys and
            values to set on the executing script.
        :return: A tuple of
            output: A unicode string of the output stream
            streams: pypsrp.powershell.PSDataStreams containing the other
                PowerShell streams
            had_errors: bool that indicates whether the script had errors
                during execution
        """
        log.info("Executing PowerShell script '%s'" % script)
        with RunspacePool(self.wsman, configuration_name=configuration_name) as pool:
            powershell = PowerShell(pool)

            if environment:
                for env_key, env_value in environment.items():
                    # Done like this for easier testing, preserves the param order
                    log.debug("Setting env var '%s' on PS script execution" % env_key)
                    powershell.add_cmdlet('New-Item'). \
                        add_parameter('Path', 'env:'). \
                        add_parameter('Name', env_key). \
                        add_parameter('Value', env_value). \
                        add_parameter('Force', True). \
                        add_cmdlet('Out-Null').add_statement()

            # so the client executes a powershell script and doesn't need to
            # deal with complex PS objects, we run the script in
            # Invoke-Expression and convert the output to a string
            # if a user wants to get the raw complex objects then they should
            # use RunspacePool and PowerShell directly
            powershell.add_cmdlet("Invoke-Expression").add_parameter("Command",
                                                                     script)
            powershell.add_cmdlet("Out-String").add_parameter("Stream")
            powershell.invoke()

        return "\n".join(powershell.output), powershell.streams, \
               powershell.had_errors

    def fetch(self, src, dest, configuration_name=DEFAULT_CONFIGURATION_NAME,
              expand_variables=False):
        """
        Will fetch a single file from the remote Windows host and create a
        local copy. Like copy(), this can be slow when it comes to fetching
        large files due to the limitation of WinRM.

        This method will first store the file in a temporary location before
        creating or replacing the file at dest if the checksum is correct.

        :param src: The path to the file on the remote host to fetch
        :param dest: The path on the localhost host to store the file as
        :param configuration_name: The PowerShell configuration endpoint to
            use when fetching the file.
        :param expand_variables: Expand variables in path. Disabled by default
            Enable for cmd like expansion (for example %TMP% in path)
        """
        if expand_variables:
            dest = os.path.expanduser(os.path.expandvars(dest))
        log.info("Fetching '%s' to '%s'" % (src, dest))

        with RunspacePool(self.wsman, configuration_name=configuration_name) as pool:
            script = get_pwsh_script('fetch.ps1')
            powershell = PowerShell(pool)
            powershell.add_script(script) \
                .add_argument(src) \
                .add_argument(expand_variables)

            log.debug("Starting remote process to output file data")
            powershell.invoke()
            _handle_powershell_error(powershell, "Failed to fetch file %s" % src)
            log.debug("Finished remote process to output file data")

            expected_hash = powershell.output[-1]

            temp_file, path = tempfile.mkstemp()
            try:
                file_bytes = base64.b64decode(powershell.output[0])
                os.write(temp_file, file_bytes)

                sha1 = hashlib.sha1()
                sha1.update(file_bytes)
                actual_hash = sha1.hexdigest()

                log.debug("Remote Hash: %s, Local Hash: %s"
                          % (expected_hash, actual_hash))
                if actual_hash != expected_hash:
                    raise WinRMError("Failed to fetch file %s, hash mismatch\n"
                                     "Source: %s\nFetched: %s"
                                     % (src, expected_hash, actual_hash))
                shutil.copy(path, dest)
            finally:
                os.close(temp_file)
                os.remove(path)

    def close(self):
        self.wsman.close()

    @staticmethod
    def sanitise_clixml(clixml):
        """
        When running a powershell script in execute_cmd (WinRS), the stderr
        stream may contain some clixml. This method will clear it up and
        replace it with the error string it would represent. This isn't done
        by default on execute_cmd for various reasons but people can call it
        manually here if they like.

        :param clixml: The clixml to parse
        :return: A unicode code string of the decoded output
        """
        output = to_unicode(clixml)
        if output.startswith("#< CLIXML"):
            # Strip off the '#< CLIXML\r\n' by finding the 2nd index of '<'
            output = output[clixml.index('<', 2):]
            element = ET.fromstring(output)
            namespace = element.tag.replace("Objs", "")[1:-1]

            errors = []
            for error in element.findall("{%s}S[@S='Error']" % namespace):
                errors.append(error.text)

            output = Serializer()._deserialize_string("".join(errors))

        return output


def _handle_powershell_error(powershell, message):
    if message and powershell.had_errors:
        errors = powershell.streams.error
        error = "\n".join([str(err) for err in errors])
        raise WinRMError("%s: %s" % (message, error))
